import json
import os
import time
from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic, Optional

import numpy as np
import torch
from pydantic import BaseModel, field_serializer, model_validator
from rlgym.api import (
    ActionSpaceType,
    ActionType,
    AgentID,
    ObsSpaceType,
    ObsType,
    RewardType,
)
from torch import nn as nn

from rlgym_learn_algos.util.torch_functions import get_device
from rlgym_learn_algos.util.torch_pydantic import (
    PydanticTorchDevice,
    PydanticTorchDtype,
)

from .actor import Actor
from .critic import Critic
from .experience_buffer import ExperienceBuffer
from .trajectory_processor import TrajectoryProcessorConfig, TrajectoryProcessorData


class PPOLearnerConfigModel(BaseModel, extra="forbid"):
    dtype: PydanticTorchDtype = torch.float32
    n_epochs: int = 1
    batch_size: int = 50000
    n_minibatches: int = 1
    ent_coef: float = 0.005
    clip_range: float = 0.2
    actor_lr: float = 3e-4
    critic_lr: float = 3e-4
    advantage_normalization: bool = True
    device: PydanticTorchDevice = "auto"
    cudnn_benchmark_mode: bool = True

    @model_validator(mode="before")
    @classmethod
    def set_device(cls, data):
        if isinstance(data, dict):
            if "device" not in data:
                data["device"] = "auto"
            data["device"] = get_device(data["device"])
        return data

    @model_validator(mode="after")
    def validate_cudnn_benchmark(self):
        if self.device.type != "cuda":
            self.cudnn_benchmark_mode = False
        return self


@dataclass
class DerivedPPOLearnerConfig:
    learner_config: PPOLearnerConfigModel
    agent_controller_name: str
    obs_space: ObsSpaceType
    action_space: ActionSpaceType
    checkpoint_load_folder: Optional[str] = None


@dataclass
class PPOData:
    batch_consumption_time: float
    cumulative_model_updates: int
    actor_entropy: float
    kl_divergence: float
    critic_loss: float
    sb3_clip_fraction: float
    actor_update_magnitude: float
    critic_update_magnitude: float


ACTOR_FILE = "actor.pt"
ACTOR_OPTIMIZER_FILE = "actor_optimizer.pt"
CRITIC_FILE = "critic.pt"
CRITIC_OPTIMIZER_FILE = "critic_optimizer.pt"
MISC_STATE = "misc.json"


class PPOLearner(
    Generic[
        TrajectoryProcessorConfig,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        ObsSpaceType,
        ActionSpaceType,
        TrajectoryProcessorData,
    ]
):
    def __init__(
        self,
        actor_factory: Callable[
            [ObsSpaceType, ActionSpaceType, torch.device],
            Actor[AgentID, ObsType, ActionType],
        ],
        critic_factory: Callable[
            [ObsSpaceType, torch.device], Critic[AgentID, ObsType]
        ],
    ):
        self.actor_factory = actor_factory
        self.critic_factory = critic_factory

    def load(self, config: DerivedPPOLearnerConfig):
        self.config = config

        if (
            config.learner_config.cudnn_benchmark_mode
            and config.learner_config.device.type == "cuda"
        ):
            torch.backends.cudnn.benchmark = True

        self.actor = self.actor_factory(
            config.obs_space, config.action_space, config.learner_config.device
        )
        self.critic = self.critic_factory(
            config.obs_space, config.learner_config.device
        )

        self.actor_optimizer = torch.optim.Adam(
            self.actor.parameters(), lr=self.config.learner_config.actor_lr
        )
        self.critic_optimizer = torch.optim.Adam(
            self.critic.parameters(), lr=self.config.learner_config.critic_lr
        )
        self.critic_loss_fn = torch.nn.MSELoss()

        # Calculate parameter counts
        actor_params = self.actor.parameters()
        critic_params = self.critic.parameters()

        trainable_actor_parameters = filter(lambda p: p.requires_grad, actor_params)
        actor_params_count = sum(p.numel() for p in trainable_actor_parameters)

        trainable_critic_parameters = filter(lambda p: p.requires_grad, critic_params)
        critic_params_count = sum(p.numel() for p in trainable_critic_parameters)

        total_parameters = actor_params_count + critic_params_count

        # Display in a structured manner
        print(f"{self.config.agent_controller_name}: Trainable Parameters:")
        print(f"{self.config.agent_controller_name}: {'Component':<10} {'Count':<10}")
        print("-" * 20)
        print(
            f"{self.config.agent_controller_name}: {'Policy':<10} {actor_params_count:<10}"
        )
        print(
            f"{self.config.agent_controller_name}: {'Critic':<10} {critic_params_count:<10}"
        )
        print("-" * 20)
        print(
            f"{self.config.agent_controller_name}: {'Total':<10} {total_parameters:<10}"
        )

        print(
            f"{self.config.agent_controller_name}: Current Policy Learning Rate: {self.config.learner_config.actor_lr}"
        )
        print(
            f"{self.config.agent_controller_name}: Current Critic Learning Rate: {self.config.learner_config.critic_lr}"
        )
        self.cumulative_model_updates = 0

        if self.config.checkpoint_load_folder is not None:
            self._load_from_checkpoint()
        self.minibatch_size = int(
            np.ceil(
                self.config.learner_config.batch_size
                / self.config.learner_config.n_minibatches
            )
        )

    def _load_from_checkpoint(self):

        assert os.path.exists(
            self.config.checkpoint_load_folder
        ), f"{self.config.agent_controller_name}: PPO Learner cannot find folder: {self.config.checkpoint_load_folder}"

        self.actor.load_state_dict(
            torch.load(
                os.path.join(self.config.checkpoint_load_folder, ACTOR_FILE),
                map_location=self.config.learner_config.device,
            )
        )
        self.critic.load_state_dict(
            torch.load(
                os.path.join(self.config.checkpoint_load_folder, CRITIC_FILE),
                map_location=self.config.learner_config.device,
            )
        )
        self.actor_optimizer.load_state_dict(
            torch.load(
                os.path.join(self.config.checkpoint_load_folder, ACTOR_OPTIMIZER_FILE),
                map_location=self.config.learner_config.device,
            )
        )
        self.critic_optimizer.load_state_dict(
            torch.load(
                os.path.join(self.config.checkpoint_load_folder, CRITIC_OPTIMIZER_FILE),
                map_location=self.config.learner_config.device,
            )
        )
        try:
            with open(
                os.path.join(self.config.checkpoint_load_folder, MISC_STATE), "rt"
            ) as f:
                misc_state = json.load(f)
                self.cumulative_model_updates = misc_state["cumulative_model_updates"]
        except FileNotFoundError:
            print(
                f"{self.config.agent_controller_name}: Tried to load the PPO learner's misc state from the file at location {str(os.path.join(self.config.checkpoint_load_folder, MISC_STATE))}, but there is no such file! Miscellaneous stats will be initialized as if this were a new run instead."
            )
            self.cumulative_model_updates = 0

    def save_checkpoint(self, folder_path):
        os.makedirs(folder_path, exist_ok=True)
        torch.save(self.actor.state_dict(), os.path.join(folder_path, ACTOR_FILE))
        torch.save(self.critic.state_dict(), os.path.join(folder_path, CRITIC_FILE))
        torch.save(
            self.actor_optimizer.state_dict(),
            os.path.join(folder_path, ACTOR_OPTIMIZER_FILE),
        )
        torch.save(
            self.critic_optimizer.state_dict(),
            os.path.join(folder_path, CRITIC_OPTIMIZER_FILE),
        )
        with open(os.path.join(folder_path, MISC_STATE), "wt") as f:
            json.dump(
                {"cumulative_model_updates": self.cumulative_model_updates}, f, indent=4
            )

    def learn(
        self,
        exp: ExperienceBuffer[
            TrajectoryProcessorConfig,
            AgentID,
            ObsType,
            ActionType,
            RewardType,
            TrajectoryProcessorData,
        ],
    ):
        """
        Compute PPO updates with an experience buffer.

        Args:
            exp (ExperienceBuffer): Experience buffer containing training data.
            collect_metrics_fn: Function to be called with the PPO metrics resulting from learn()
        """

        n_batches = 0
        clip_fractions = []
        entropies = []
        divergences = []
        val_losses = []

        # Save parameters before computing any updates.
        actor_before = torch.nn.utils.parameters_to_vector(self.actor.parameters())
        critic_before = torch.nn.utils.parameters_to_vector(self.critic.parameters())

        t1 = time.time()
        for epoch in range(self.config.learner_config.n_epochs):
            # Get all shuffled batches from the experience buffer.
            batches = exp.get_all_batches_shuffled(
                self.config.learner_config.batch_size
            )
            for batch in batches:
                (
                    batch_agent_ids,
                    batch_obs,
                    batch_acts,
                    batch_old_probs,
                    batch_values,
                    batch_advantages,
                ) = batch
                batch_target_values = batch_values + batch_advantages
                if self.config.learner_config.advantage_normalization:
                    old_device = batch_advantages.device
                    batch_advantages = batch_advantages.to(
                        self.config.learner_config.device
                    )
                    std, mean = torch.std_mean(batch_advantages)
                    batch_advantages = (batch_advantages - mean) / (std + 1e-8)
                    batch_advantages = batch_advantages.to(old_device)

                self.actor_optimizer.zero_grad()
                self.critic_optimizer.zero_grad()

                for minibatch_slice in range(
                    0, self.config.learner_config.batch_size, self.minibatch_size
                ):
                    # Send everything to the device and enforce correct shapes.
                    start = minibatch_slice
                    stop = min(
                        start + self.minibatch_size,
                        self.config.learner_config.batch_size,
                    )
                    minibatch_ratio = (
                        stop - start
                    ) / self.config.learner_config.batch_size

                    agent_ids = batch_agent_ids[start:stop]
                    obs = batch_obs[start:stop]
                    acts = batch_acts[start:stop]
                    advantages = batch_advantages[start:stop].to(
                        self.config.learner_config.device
                    )
                    old_probs = batch_old_probs[start:stop].to(
                        self.config.learner_config.device
                    )
                    target_values = batch_target_values[start:stop].to(
                        self.config.learner_config.device
                    )

                    # Compute value estimates.
                    vals = self.critic(agent_ids, obs).view_as(target_values)

                    # Get actor log probs & entropy.
                    log_probs, entropy = self.actor.get_backprop_data(
                        agent_ids, obs, acts
                    )
                    log_probs = log_probs.view_as(old_probs)
                    entropy = entropy * minibatch_ratio

                    # Compute PPO loss.
                    ratio = torch.exp(log_probs - old_probs)
                    clipped = torch.clamp(
                        ratio,
                        1.0 - self.config.learner_config.clip_range,
                        1.0 + self.config.learner_config.clip_range,
                    )

                    # Compute KL divergence & clip fraction using SB3 method for reporting.
                    with torch.no_grad():
                        log_ratio = log_probs - old_probs
                        kl = (torch.exp(log_ratio) - 1) - log_ratio
                        kl = kl.mean().detach() * minibatch_ratio

                        # From the stable-baselines3 implementation of PPO.
                        clip_fraction = torch.mean(
                            (
                                torch.abs(ratio - 1)
                                > self.config.learner_config.clip_range
                            ).float()
                        ).to(device="cpu", non_blocking=True)
                        clip_fractions.append((clip_fraction, minibatch_ratio))

                    actor_loss = (
                        -torch.min(ratio * advantages, clipped * advantages).mean()
                        * minibatch_ratio
                    )
                    value_loss = (
                        self.critic_loss_fn(vals, target_values) * minibatch_ratio
                    )
                    ppo_loss = (
                        actor_loss - entropy * self.config.learner_config.ent_coef
                    )

                    ppo_loss.backward()
                    value_loss.backward()

                    val_losses.append(
                        value_loss.to(device="cpu", non_blocking=True).detach()
                    )
                    divergences.append(kl.to(device="cpu", non_blocking=True).detach())
                    entropies.append(
                        entropy.to(device="cpu", non_blocking=True).detach()
                    )

                torch.nn.utils.clip_grad_norm_(self.critic.parameters(), max_norm=0.5)
                torch.nn.utils.clip_grad_norm_(self.actor.parameters(), max_norm=0.5)

                self.actor_optimizer.step()
                self.critic_optimizer.step()

                n_batches += 1

        # Compute magnitude of updates made to the actor and critic.
        actor_after = torch.nn.utils.parameters_to_vector(self.actor.parameters())
        critic_after = torch.nn.utils.parameters_to_vector(self.critic.parameters())
        actor_update_magnitude = (actor_before - actor_after).norm().cpu().item()
        critic_update_magnitude = (critic_before - critic_after).norm().cpu().item()

        if self.config.learner_config.device.type != "cpu":
            torch.cuda.current_stream().synchronize()

        tot_clip = sum(
            v.item() * minibatch_ratio for (v, minibatch_ratio) in clip_fractions
        )
        tot_entropy = sum(v.item() for v in entropies)
        tot_divergence = sum(v.item() for v in divergences)
        tot_val_loss = sum(v.item() for v in val_losses)

        self.actor_optimizer.zero_grad()
        self.critic_optimizer.zero_grad()

        self.cumulative_model_updates += n_batches

        # If there were no batches, we just want to log the total time spent here (and totals will all be 0 anyway), so just set n_batches to 1
        if n_batches == 0:
            n_batches = 1
        mean_clip = tot_clip / n_batches
        mean_entropy = tot_entropy / n_batches
        mean_divergence = tot_divergence / n_batches
        mean_val_loss = tot_val_loss / n_batches
        return PPOData(
            (time.time() - t1) / n_batches,
            self.cumulative_model_updates,
            mean_entropy,
            mean_divergence,
            mean_val_loss,
            mean_clip,
            actor_update_magnitude,
            critic_update_magnitude,
        )



from abc import abstractmethod
from typing import Generic, List

import torch.nn as nn
from rlgym.api import AgentID, ObsType
from torch import Tensor


class Critic(nn.Module, Generic[AgentID, ObsType]):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def forward(self, agent_id_list: List[AgentID], obs_list: List[ObsType]) -> Tensor:
        """
        :param agent_id_list: List of AgentIDs, parallel with obs_list. AgentIDs may not be unique here.
        :param obs_list: List of ObsTypes to compute values for.
        :return: Tensor. Must be 0-dimensional for PPO, with dtype float32.
        """
        raise NotImplementedError



from typing import List, Tuple

import numpy as np
from rlgym.api import AgentID

from rlgym_learn_algos.util.running_stats import WelfordRunningStat

from .obs_standardizer import ObsStandardizer


class NumpyObsStandardizer(ObsStandardizer[AgentID, np.ndarray]):
    def __init__(
        self, steps_per_obs_stats_update: int, steps_until_fixed: int = np.inf
    ):
        self.obs_stats = None
        self.steps_per_obs_stats_update = steps_per_obs_stats_update
        self.obs_stats_start_index = 0
        self.steps_until_fixed = steps_until_fixed
        self.steps = 0

    def standardize(self, obs_list) -> List[Tuple[AgentID, np.ndarray]]:
        if self.obs_stats == None:
            (_, obs) = obs_list[0]
            self.obs_stats = WelfordRunningStat(obs.shape)
        if self.steps < self.steps_until_fixed:
            stats_update_batch = [
                o[1]
                for o in obs_list[
                    self.obs_stats_start_index :: self.steps_per_obs_stats_update
                ]
            ]
            self.obs_stats_start_index = (
                self.steps_per_obs_stats_update
                - 1
                - (
                    (len(obs_list) - self.obs_stats_start_index - 1)
                    % self.steps_per_obs_stats_update
                )
            )
            for sample in stats_update_batch:
                self.obs_stats.update(sample)
        return [
            (agent_id, (obs - self.obs_stats.mean) / self.obs_stats.std)
            for (agent_id, obs) in obs_list
        ]



from abc import abstractmethod
from typing import Generic, Iterable, List, Tuple

import torch.nn as nn
from rlgym.api import ActionType, AgentID, ObsType
from torch import Tensor


class Actor(nn.Module, Generic[AgentID, ObsType, ActionType]):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_action(
        self, agent_id_list: List[AgentID], obs_list: List[ObsType], **kwargs
    ) -> Tuple[Iterable[ActionType], Tensor]:
        """
        Function to get an action and the log of its probability from the policy given an observation.
        :param agent_id_list: List of AgentIDs for which to produce actions. AgentIDs may not be unique here. Parallel with obs_list.
        :param obs_list: List of ObsTypes for which to produce actions. Parallel with agent_id_list.
        :return: Tuple of a list of chosen actions and Tensor with shape (n,) of log probs (float32), with the action list and the first (only) dimension of the tensor parallel with obs_list.
        """
        raise NotImplementedError

    @abstractmethod
    def get_backprop_data(
        self,
        agent_id_list: List[AgentID],
        obs_list: List[ObsType],
        acts: List[ActionType],
        **kwargs,
    ) -> Tuple[Tensor, Tensor]:
        """
        Function to compute the data necessary for backpropagation.
        :param agent_id_list: list of agent ids, parallel with obs_list.
        :param obs_list: list of ObsTypes to pass through the policy
        :param acts: Actions taken by the policy, parallel with obs_list
        :return: (Action log probs tensor with first dimension parallel with acts, mean entropy as 0-dimensional tensor).
        """
        raise NotImplementedError



from abc import abstractmethod
from typing import Generic, List

from rlgym.api import AgentID, ObsType


class ObsStandardizer(Generic[AgentID, ObsType]):
    @abstractmethod
    def standardize(
        self, agent_id_list: List[AgentID], obs_list: List[ObsType]
    ) -> List[ObsType]:
        """
        :param agent_id_list: List of AgentIDs, parallel with obs_list. AgentIDs may not be unique here.
        :param obs_list: List of ObsTypes to standardize.
        :return: List of standardized observations, parallel with input lists.
        """
        raise NotImplementedError



from typing import Any, Dict, List

from rlgym_learn_algos.logging import DictMetricsLogger
from rlgym_learn_algos.ppo import PPOAgentControllerData

from .gae_trajectory_processor import GAETrajectoryProcessorData


class PPOMetricsLogger(
    DictMetricsLogger[
        None,
        None,
        PPOAgentControllerData[GAETrajectoryProcessorData],
    ],
):
    def __init__(self):
        self.state_metrics: Dict[str, Any] = {}
        self.agent_metrics: Dict[str, Any] = {}

    def get_metrics(self) -> Dict[str, Any]:
        return {**self.agent_metrics, **self.state_metrics}

    def collect_env_metrics(self, data: List[Dict[str, Any]]):
        """
        Override this function to set self.state_metrics to something else using the data provided.
        The metrics should be nested dictionaries
        """
        self.state_metrics = {}

    def collect_agent_metrics(
        self, data: PPOAgentControllerData[GAETrajectoryProcessorData]
    ):
        self.agent_metrics = {
            "Timing": {
                "PPO Batch Consumption Time": data.ppo_data.batch_consumption_time,
                "Total Iteration Time": data.iteration_time,
                "Timestep Collection Time": data.timestep_collection_time,
                "Timestep Consumption Time": data.iteration_time
                - data.timestep_collection_time,
                "Collected Steps per Second": data.timesteps_collected
                / data.timestep_collection_time,
                "Overall Steps per Second": data.timesteps_collected
                / data.iteration_time,
            },
            "Timestep Collection": {
                "Cumulative Timesteps": data.cumulative_timesteps,
                "Timesteps Collected": data.timesteps_collected,
            },
            "PPO Metrics": {
                "Average Reward": data.trajectory_processor_data.average_reward,
                "Average Undiscounted Episodic Return": data.trajectory_processor_data.average_undiscounted_episodic_return,
                "Average Return": data.trajectory_processor_data.average_return,
                "Return Standard Deviation": data.trajectory_processor_data.return_standard_deviation,
                "Cumulative Model Updates": data.ppo_data.cumulative_model_updates,
                "Actor Entropy": data.ppo_data.actor_entropy,
                "Mean KL Divergence": data.ppo_data.kl_divergence,
                "Critic Loss": data.ppo_data.critic_loss,
                "SB3 Clip Fraction": data.ppo_data.sb3_clip_fraction,
                "Actor Update Magnitude": data.ppo_data.actor_update_magnitude,
                "Critic Update Magnitude": data.ppo_data.critic_update_magnitude,
            },
        }

    def validate_config(self, config_obj) -> None:
        return None



"""
File: running_stats.py
Author: Matthew Allen

Description:
    An implementation of Welford's algorithm for running statistics.
"""

import numpy as np


class WelfordRunningStat(object):
    """
    https://www.johndcook.com/blog/skewness_kurtosis/
    """

    def __init__(self, shape):
        self.ones = np.ones(shape=shape, dtype=np.float32)
        self.zeros = np.zeros(shape=shape, dtype=np.float32)

        self.running_mean = np.zeros(shape=shape, dtype=np.float32)
        self.running_variance = np.zeros(shape=shape, dtype=np.float32)

        self.count = 0
        self.shape = shape

    def increment(self, samples, num):
        if num > 1:
            for i in range(num):
                self.update(samples[i])
        else:
            self.update(samples)

    def update(self, sample):
        if type(sample) == dict:
            sample = sample["frame"]
        current_count = self.count
        self.count += 1
        delta = (sample - self.running_mean).reshape(self.running_mean.shape)
        delta_n = (delta / self.count).reshape(self.running_mean.shape)

        self.running_mean += delta_n
        self.running_variance += delta * delta_n * current_count

    def reset(self):
        del self.running_mean
        del self.running_variance

        self.__init__(self.shape)

    @property
    def mean(self):
        if self.count < 2:
            return self.zeros
        return self.running_mean

    @property
    def std(self):
        if self.count < 2:
            return self.ones

        var = self.running_variance / (self.count - 1)

        # Wherever variance is zero, set it to 1 to avoid division by zero.
        var = np.where(var == 0, 1.0, var)
        return np.sqrt(var)

    def increment_from_serialized_other(self, serialized_other):
        """
        Function to combine the statistics computed by two independent instances of this algorithm.
        :param serialized_other: A list containing the serialization of the other instance to be combined with.
        :return: None.
        """
        # Unpack the serialized data.
        n = np.prod(self.shape)
        other_mean = np.asarray(serialized_other[:n], dtype=np.float32).reshape(
            self.running_mean.shape
        )
        other_var = np.asarray(serialized_other[n:-1], dtype=np.float32).reshape(
            self.running_variance.shape
        )
        other_count = serialized_other[-1]
        if other_count == 0:
            return

        # Combine our statistics with the new data.
        count = self.count + other_count

        mean_delta = other_mean - self.running_mean
        mean_delta_squared = mean_delta * mean_delta

        combined_mean = (
            self.count * self.running_mean + other_count * other_mean
        ) / count

        combined_variance = (
            self.running_variance
            + other_var
            + mean_delta_squared * self.count * other_count / count
        )

        # Update our local variables to the newly combined statistics.
        self.running_mean = combined_mean
        self.running_variance = combined_variance
        self.count = count

    def serialize(self):
        return (
            self.running_mean.ravel().tolist()
            + self.running_variance.ravel().tolist()
            + [self.count]
        )

    def deserialize(self, other):
        self.reset()
        n = np.prod(self.shape)

        other_mean = other[:n]
        other_var = other[n:-1]
        other_count = other[-1]
        self.running_mean = np.reshape(other_mean, self.shape)
        self.running_variance = np.reshape(other_var, self.shape)
        self.count = other_count

    def state_dict(self):
        return {
            "mean": self.running_mean.ravel().tolist(),
            "var": self.running_variance.ravel().tolist(),
            "shape": np.shape(self.running_mean),
            "count": self.count,
        }

    def load_state_dict(self, state):
        shape = state["shape"]
        self.count = state["count"]
        self.running_mean = np.asarray(state["mean"]).reshape(shape)
        self.running_variance = np.asarray(state["var"]).reshape(shape)
        print(
            f"LOADED RUNNING STATS FROM JSON | Mean: {self.running_mean} | Variance: {self.running_variance} | Count: {self.count}"
        )



import numpy as np
import torch
import torch.nn as nn
from rlgym.api import AgentID

from .critic import Critic


class BasicCritic(Critic[AgentID, np.ndarray]):
    def __init__(self, input_size, layer_sizes, device):
        super().__init__()
        self.device = device

        assert (
            len(layer_sizes) != 0
        ), "AT LEAST ONE LAYER MUST BE SPECIFIED TO BUILD THE NEURAL NETWORK!"
        layers = [nn.Linear(input_size, layer_sizes[0]), nn.ReLU()]

        prev_size = layer_sizes[0]
        for size in layer_sizes[1:]:
            layers.append(nn.Linear(prev_size, size))
            layers.append(nn.ReLU())
            prev_size = size

        layers.append(nn.Linear(layer_sizes[-1], 1))
        self.model = nn.Sequential(*layers).to(self.device)

    def forward(self, agent_id_list, obs_list) -> torch.Tensor:
        obs = torch.as_tensor(
            np.array(obs_list), dtype=torch.float32, device=self.device
        )
        return self.model(obs)



from abc import abstractmethod
from typing import Any, Generic, List

import numpy as np
from numpy import dtype, ndarray
from rlgym.api import RewardType


class BatchRewardTypeNumpyConverter(Generic[RewardType]):
    @abstractmethod
    def as_numpy(self, rewards: List[RewardType]):
        """
        :param rewards: A list of RewardType to be converted
        :return: A Numpy array of shape (n,) parallel to the input list, with dtype self.dtype
        """
        raise NotImplementedError

    def set_dtype(self, dtype: dtype):
        self.dtype = dtype


class BatchRewardTypeSimpleNumpyConverter(BatchRewardTypeNumpyConverter[Any]):
    def as_numpy(self, rewards) -> ndarray:
        return np.array(rewards, dtype=self.dtype)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, TypeVar

from rlgym_learn.api import AgentControllerData

MetricsLoggerConfig = TypeVar("MetricsLoggerConfig")
MetricsLoggerAdditionalDerivedConfig = TypeVar("MetricsLoggerAdditionalDerivedConfig")


@dataclass
class DerivedMetricsLoggerConfig(
    Generic[MetricsLoggerConfig, MetricsLoggerAdditionalDerivedConfig]
):
    metrics_logger_config: MetricsLoggerConfig = None
    checkpoint_load_folder: Optional[str] = None
    agent_controller_name: str = ""
    additional_derived_config: MetricsLoggerAdditionalDerivedConfig = None


# TODO: docs
class MetricsLogger(
    Generic[
        MetricsLoggerConfig,
        MetricsLoggerAdditionalDerivedConfig,
        AgentControllerData,
    ]
):
    """
    This class is designed to be used inside an agent controller to handle the processing of state metrics and agent controller data, and to have some side effects resulting from said processing. It supports config-based saving and loading, and nesting with other MetricsLogger subclasses' config-based saving and loading via the AdditionalDerivedConfig.

    MetricsLoggerConfig is the (pydantic) config model for the class, or None if no config is needed.

    MetricsLoggerAdditionalConfig is a dataclass that can include arbitrary data (usually from other config models). It is the responsibility of the agent controller to instantiate this if it's needed.

    StateMetrics is the type used for collection of data from the environment processes.

    AgentControllerData is the type used for collection of data from the agent controller containing this metrics logger.
    """

    def collect_env_metrics(self, data: List[Dict[str, Any]]):
        """
        This method is intended to allow batch processing of env metrics using the shared info deserialized from the env processes. The result of processing should be stored and used the next time report_metrics is called.
        There is no guarantee that this method will only be called once between each report_metrics call.
        """
        pass

    def collect_agent_metrics(self, data: AgentControllerData):
        """
        This method is intended to allow processing of AgentControllerData after it gets finalized by the agent controller. The result of processing should be stored and used the next time report_metrics is called.
        There is no guarantee that this method will only be called once between each report_metrics call.
        """
        pass

    @abstractmethod
    def report_metrics(self):
        """
        This method is intended to have arbitrary side effects based on data collected so far. This could be printing, or logging to wandb, or sending data to a redis server, or whatever.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_config(self, config_obj: Dict[str, Any]) -> MetricsLoggerConfig:
        """
        Any class inheriting from this one has some sort of pydantic config (or None) that it expects to receive for loading. The agent controller may be generic over the type of metrics logger used, so it
        needs some way of creating an instance of the particular MetricsLoggerConfig used based on a general-purpose config dict that doesn't have any guarantees on contents.

        :return: an instance of the (pydantic) config model used by this class (or None), to be placed inside an instance of DerivedMetricsLoggerConfig when calling load.
        """
        raise NotImplementedError

    def load(
        self,
        config: DerivedMetricsLoggerConfig[
            MetricsLoggerConfig, MetricsLoggerAdditionalDerivedConfig
        ],
    ):
        """
        Sets data inside this instance using config, which may include loading data from a checkpoint.
        """
        pass

    def save_checkpoint(self, folder_path, file_name):
        """
        Saves data inside this instance which needs to be checkpointed.
        """
        pass



import json
import os
import time
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, TypeVar

import wandb
from pydantic import BaseModel, Field
from rlgym_learn.api import AgentControllerData

from .dict_metrics_logger import DictMetricsLogger
from .metrics_logger import DerivedMetricsLoggerConfig, MetricsLogger

InnerMetricsLoggerConfig = TypeVar("InnerMetricsLoggerConfig")
InnerMetricsLoggerAdditionalDerivedConfig = TypeVar(
    "InnerMetricsLoggerAdditionalDerivedConfig"
)


def convert_nested_dict(d):
    new = {}
    for k, v in d.items():
        if isinstance(v, dict):
            converted = convert_nested_dict(v)
            to_add = {f"{k}/{k1}": v1 for k1, v1 in converted.items()}
        else:
            to_add = {k: v}
        new = {**new, **to_add}
    return new


class WandbMetricsLoggerConfigModel(BaseModel, extra="forbid"):
    enable: bool = True
    project: str = "rlgym-learn"
    group: str = "unnamed-runs"
    run: str = "rlgym-learn-run"
    id: Optional[str] = None
    new_run_with_timestamp_suffix: bool = False
    additional_wandb_run_config: Dict[str, Any] = Field(default_factory=dict)
    settings_kwargs: Dict[str, Any] = Field(default_factory=dict)


@dataclass
class WandbAdditionalDerivedConfig(
    Generic[InnerMetricsLoggerConfig, InnerMetricsLoggerAdditionalDerivedConfig]
):
    derived_wandb_run_config: Dict[str, Any] = Field(default_factory=dict)
    timestamp_suffix: Optional[str] = None
    inner_metrics_logger_config: InnerMetricsLoggerConfig = None
    inner_metrics_logger_additional_derived_config: (
        InnerMetricsLoggerAdditionalDerivedConfig
    ) = None


class WandbMetricsLogger(
    MetricsLogger[
        WandbMetricsLoggerConfigModel,
        WandbAdditionalDerivedConfig[
            InnerMetricsLoggerConfig, InnerMetricsLoggerAdditionalDerivedConfig
        ],
        AgentControllerData,
    ],
    Generic[
        InnerMetricsLoggerConfig,
        InnerMetricsLoggerAdditionalDerivedConfig,
        AgentControllerData,
    ],
):
    def __init__(
        self,
        inner_metrics_logger: DictMetricsLogger[
            InnerMetricsLoggerConfig,
            InnerMetricsLoggerAdditionalDerivedConfig,
            AgentControllerData,
        ],
        checkpoint_file_name: str = "wandb_metrics_logger.json",
    ):
        self.inner_metrics_logger = inner_metrics_logger
        self.checkpoint_file_name = checkpoint_file_name
        self.run_id = None

    def collect_env_metrics(self, data: List[Dict[str, Any]]):
        self.inner_metrics_logger.collect_env_metrics(data)

    def collect_agent_metrics(self, data: AgentControllerData):
        self.inner_metrics_logger.collect_agent_metrics(data)

    def report_metrics(self):
        self.wandb_run.log(convert_nested_dict(self.inner_metrics_logger.get_metrics()))
        self.inner_metrics_logger.report_metrics()

    def validate_config(self, config_obj: Any):
        return WandbMetricsLoggerConfigModel.model_validate(config_obj)

    def load(self, config):
        self.config = config
        self.inner_metrics_logger.load(
            DerivedMetricsLoggerConfig(
                checkpoint_load_folder=config.checkpoint_load_folder,
                agent_controller_name=config.agent_controller_name,
                metrics_logger_config=config.additional_derived_config.inner_metrics_logger_config,
                additional_derived_config=config.additional_derived_config.inner_metrics_logger_additional_derived_config,
            )
        )
        if self.config.checkpoint_load_folder is not None:
            self._load_from_checkpoint()
        if not self.config.metrics_logger_config.enable:
            self.wandb_run = None
            self.run_id = None
            return

        if self.run_id is not None and self.config.metrics_logger_config.id is not None:
            print(
                f"{self.config.agent_controller_name}: Wandb run id from checkpoint ({self.run_id}) is being overridden by wandb run id from config: {self.config.metrics_logger_config.id}"
            )
            self.run_id = self.config.metrics_logger_config.id

        wandb_config = {
            **self.config.additional_derived_config.derived_wandb_run_config,
            **self.config.metrics_logger_config.additional_wandb_run_config,
        }

        run_name = self.config.metrics_logger_config.run
        if self.config.metrics_logger_config.new_run_with_timestamp_suffix:
            print(
                f"{self.config.agent_controller_name}: Due to config, a new wandb run is being created with timestamp suffix. This run will use the project and group specified in config, and will use the run name in config prepended to the timestamp suffix."
            )
            if (
                self.config.additional_derived_config.timestamp_suffix is not None
                and len(self.config.additional_derived_config.timestamp_suffix) > 0
            ):
                run_name += self.config.additional_derived_config.timestamp_suffix
            else:
                run_name += f"-{time.time_ns()}"

        self.wandb_run = wandb.init(
            project=self.config.metrics_logger_config.project,
            group=self.config.metrics_logger_config.group,
            config=wandb_config,
            name=run_name,
            id=self.run_id,
            resume="allow",
            reinit=True,
            settings=wandb.Settings(
                **self.config.metrics_logger_config.settings_kwargs
            ),
        )
        self.run_id = self.wandb_run.id
        print(f"{self.config.agent_controller_name}: Created wandb run! {self.run_id}")

    def _load_from_checkpoint(self):
        try:
            with open(
                os.path.join(
                    self.config.checkpoint_load_folder,
                    self.checkpoint_file_name,
                ),
                "rt",
            ) as f:
                state = json.load(f)
            if "run_id" in state:
                self.run_id = state["run_id"]
            else:
                self.run_id = None
        except FileNotFoundError:
            print(
                f"{self.config.agent_controller_name}: Tried to load wandb run from checkpoint using the file at location {str(os.path.join(self.config.checkpoint_load_folder, self.checkpoint_file_name))}, but there is no such file! A new run will be created based on the config values instead."
            )
            self.run_id = None

    def save_checkpoint(self, folder_path):
        os.makedirs(folder_path, exist_ok=True)
        state = {"run_id": self.run_id}
        with open(
            os.path.join(
                folder_path,
                self.checkpoint_file_name,
            ),
            "wt",
        ) as f:
            json.dump(state, f, indent=4)



from .dict_metrics_logger import DictMetricsLogger
from .metrics_logger import (
    DerivedMetricsLoggerConfig,
    MetricsLogger,
    MetricsLoggerAdditionalDerivedConfig,
    MetricsLoggerConfig,
)
from .wandb_metrics_logger import (
    InnerMetricsLoggerAdditionalDerivedConfig,
    InnerMetricsLoggerConfig,
    WandbAdditionalDerivedConfig,
    WandbMetricsLogger,
    WandbMetricsLoggerConfigModel,
)



"""
File: discrete_policy.py
Author: Matthew Allen

Description:
    An implementation of a feed-forward neural network which parametrizes a discrete distribution over a space of actions.
"""

from typing import Iterable, List, Tuple

import numpy as np
import torch
import torch.nn as nn
from rlgym.api import AgentID
from torch.distributions.utils import probs_to_logits

from .actor import Actor


class DiscreteFF(Actor[AgentID, np.ndarray, np.ndarray]):
    def __init__(self, input_size, n_actions, layer_sizes, device):
        super().__init__()
        self.device = device

        assert (
            len(layer_sizes) != 0
        ), "AT LEAST ONE LAYER MUST BE SPECIFIED TO BUILD THE NEURAL NETWORK!"
        layers = [nn.Linear(input_size, layer_sizes[0]), nn.ReLU()]
        prev_size = layer_sizes[0]
        for size in layer_sizes[1:]:
            layers.append(nn.Linear(prev_size, size))
            layers.append(nn.ReLU())
            prev_size = size

        layers.append(nn.Linear(layer_sizes[-1], n_actions))
        layers.append(nn.Softmax(dim=-1))
        self.model = nn.Sequential(*layers).to(self.device)

        self.n_actions = n_actions

    def get_output(self, obs_list: List[np.ndarray]) -> torch.Tensor:
        obs = torch.as_tensor(
            np.array(obs_list), dtype=torch.float32, device=self.device
        )
        probs = self.model(obs)
        probs = torch.clamp(probs, min=1e-11, max=1)
        return probs

    def get_action(
        self, agent_id_list, obs_list, **kwargs
    ) -> Tuple[Iterable[np.ndarray], torch.Tensor]:
        probs = self.get_output(obs_list)
        if "deterministic" in kwargs and kwargs["deterministic"]:
            action = probs.cpu().numpy().argmax(axis=-1)
            return action, torch.zeros(action.shape)

        action = torch.multinomial(probs, 1, True)
        log_prob: torch.Tensor = torch.log(probs).gather(-1, action)

        return action.cpu().numpy(), log_prob.squeeze().to(
            device="cpu", non_blocking=True
        )

    def get_backprop_data(self, agent_id_list, obs_list, acts, **kwargs):
        probs = self.get_output(obs_list)
        acts_tensor = torch.as_tensor(np.array(acts)).to(self.device)
        logits = probs_to_logits(probs)
        min_real = torch.finfo(logits.dtype).min
        logits = torch.clamp(logits, min=min_real)
        entropy = -(logits * probs).sum(dim=-1)
        action_logits = logits.gather(-1, acts_tensor)

        return action_logits.to(self.device), entropy.to(self.device).mean()



from __future__ import annotations

import json
import os
import pickle
import random
import shutil
import time
from collections.abc import Callable
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, Tuple, Union

import numpy as np
import torch
from pydantic import BaseModel, Field, model_validator
from rlgym.api import (
    ActionSpaceType,
    ActionType,
    AgentID,
    ObsSpaceType,
    ObsType,
    RewardType,
    StateType,
)
from rlgym_learn import EnvActionResponse, EnvActionResponseType, Timestep
from rlgym_learn.api.agent_controller import AgentController
from torch import device as _device

from rlgym_learn_algos.logging import (
    DerivedMetricsLoggerConfig,
    MetricsLogger,
    MetricsLoggerAdditionalDerivedConfig,
    MetricsLoggerConfig,
    WandbAdditionalDerivedConfig,
    WandbMetricsLogger,
)
from rlgym_learn_algos.stateful_functions import ObsStandardizer
from rlgym_learn_algos.util.torch_functions import get_device

from .actor import Actor
from .critic import Critic
from .env_trajectories import EnvTrajectories
from .experience_buffer import (
    DerivedExperienceBufferConfig,
    ExperienceBuffer,
    ExperienceBufferConfigModel,
)
from .ppo_learner import (
    DerivedPPOLearnerConfig,
    PPOData,
    PPOLearner,
    PPOLearnerConfigModel,
)
from .trajectory import Trajectory
from .trajectory_processor import TrajectoryProcessorConfig, TrajectoryProcessorData

EXPERIENCE_BUFFER_FOLDER = "experience_buffer"
PPO_LEARNER_FOLDER = "ppo_learner"
METRICS_LOGGER_FOLDER = "metrics_logger"
PPO_AGENT_FILE = "ppo_agent.json"
ITERATION_TRAJECTORIES_FILE = "current_trajectories.pkl"  # this should be renamed, but it would be a breaking change so I'm leaving it until I happen to make one of those and remember to update this at the same time
ITERATION_SHARED_INFOS_FILE = "iteration_shared_infos.pkl"


class PPOAgentControllerConfigModel(BaseModel, extra="forbid"):
    timesteps_per_iteration: int = 50000
    save_every_ts: int = 1_000_000
    add_unix_timestamp: bool = True
    checkpoint_load_folder: Optional[str] = None
    n_checkpoints_to_keep: int = 5
    random_seed: int = 123
    save_mid_iteration_data_in_checkpoint: bool = True
    learner_config: PPOLearnerConfigModel = Field(default_factory=PPOLearnerConfigModel)
    experience_buffer_config: ExperienceBufferConfigModel = Field(
        default_factory=ExperienceBufferConfigModel
    )
    run_name: str = "rlgym-learn-run"
    metrics_logger_config: Dict[str, Any] = Field(default_factory=dict)

    @model_validator(mode="before")
    @classmethod
    def set_metrics_logger_config(cls, data):
        if isinstance(data, PPOAgentControllerConfigModel):
            if isinstance(data.metrics_logger_config, BaseModel):
                data.metrics_logger_config = data.metrics_logger_config.model_dump()
        elif isinstance(data, dict) and "metrics_logger_config" in data:
            if isinstance(data["metrics_logger_config"], BaseModel):
                data["metrics_logger_config"] = data[
                    "metrics_logger_config"
                ].model_dump()
        return data


@dataclass
class PPOAgentControllerData(Generic[TrajectoryProcessorData]):
    ppo_data: PPOData
    trajectory_processor_data: TrajectoryProcessorData
    cumulative_timesteps: int
    iteration_time: float
    timesteps_collected: int
    timestep_collection_time: float


class PPOAgentController(
    AgentController[
        PPOAgentControllerConfigModel,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        StateType,
        ObsSpaceType,
        ActionSpaceType,
        torch.Tensor,
        PPOAgentControllerData[TrajectoryProcessorData],
    ],
    Generic[
        TrajectoryProcessorConfig,
        MetricsLoggerConfig,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        StateType,
        ObsSpaceType,
        ActionSpaceType,
        TrajectoryProcessorData,
    ],
):
    def __init__(
        self,
        actor_factory: Callable[
            [ObsSpaceType, ActionSpaceType, _device],
            Actor[AgentID, ObsType, ActionType],
        ],
        critic_factory: Callable[[ObsSpaceType, _device], Critic[AgentID, ObsType]],
        experience_buffer: ExperienceBuffer[
            TrajectoryProcessorConfig,
            AgentID,
            ObsType,
            ActionType,
            RewardType,
            TrajectoryProcessorData,
        ],
        metrics_logger: Optional[
            MetricsLogger[
                MetricsLoggerConfig,
                MetricsLoggerAdditionalDerivedConfig,
                PPOAgentControllerData[TrajectoryProcessorData],
            ]
        ] = None,
        obs_standardizer: Optional[ObsStandardizer] = None,
        agent_choice_fn: Callable[
            [List[AgentID]], List[int]
        ] = lambda agent_id_list: list(range(len(agent_id_list))),
    ):
        self.learner = PPOLearner(actor_factory, critic_factory)
        self.experience_buffer = experience_buffer
        self.metrics_logger = metrics_logger
        self.obs_standardizer = obs_standardizer
        if obs_standardizer is not None:
            print(
                "Warning: using an obs standardizer is slow! It is recommended to design your obs to be standardized (i.e. have approximately mean 0 and std 1 for each value) without needing this extra post-processing step."
            )
        self.agent_choice_fn = agent_choice_fn

        self.current_env_trajectories: Dict[
            str,
            EnvTrajectories[AgentID, ActionType, ObsType, RewardType],
        ] = {}
        self.iteration_trajectories: List[
            Trajectory[AgentID, ActionType, ObsType, RewardType]
        ] = []
        self.iteration_shared_infos: List[Dict[str, Any]] = []
        self.cur_iteration = 0
        self.iteration_timesteps = 0
        self.cumulative_timesteps = 0
        cur_time = time.perf_counter()
        self.iteration_start_time = cur_time
        self.timestep_collection_start_time = cur_time
        self.ts_since_last_save = 0

    def set_space_types(self, obs_space, action_space):
        self.obs_space = obs_space
        self.action_space = action_space

    def validate_config(self, config_obj):
        return PPOAgentControllerConfigModel.model_validate(config_obj)

    def load(self, config):
        self.config = config
        print(
            f"{self.config.agent_controller_name}: Using device {config.agent_controller_config.learner_config.device}"
        )
        agent_controller_config = config.agent_controller_config
        learner_config = config.agent_controller_config.learner_config
        experience_buffer_config = (
            config.agent_controller_config.experience_buffer_config
        )
        learner_checkpoint_load_folder = (
            None
            if agent_controller_config.checkpoint_load_folder is None
            else os.path.join(
                agent_controller_config.checkpoint_load_folder, PPO_LEARNER_FOLDER
            )
        )
        experience_buffer_checkpoint_load_folder = (
            None
            if agent_controller_config.checkpoint_load_folder is None
            else os.path.join(
                agent_controller_config.checkpoint_load_folder, EXPERIENCE_BUFFER_FOLDER
            )
        )
        metrics_logger_checkpoint_load_folder = (
            None
            if agent_controller_config.checkpoint_load_folder is None
            else os.path.join(
                agent_controller_config.checkpoint_load_folder, METRICS_LOGGER_FOLDER
            )
        )

        run_suffix = (
            f"-{time.time_ns()}" if agent_controller_config.add_unix_timestamp else ""
        )

        if agent_controller_config.checkpoint_load_folder is not None:
            loaded_checkpoint_runs_folder = os.path.abspath(
                os.path.join(agent_controller_config.checkpoint_load_folder, "../..")
            )
            abs_save_folder = os.path.abspath(config.save_folder)
            # TODO: this doesn't seem to be working
            if abs_save_folder == loaded_checkpoint_runs_folder:
                print(
                    f"{config.agent_controller_name}: Using the loaded checkpoint's run folder as the checkpoints save folder."
                )
                checkpoints_save_folder = os.path.abspath(
                    os.path.join(agent_controller_config.checkpoint_load_folder, "..")
                )
            else:
                print(
                    f"{config.agent_controller_name}: Runs folder in config does not align with loaded checkpoint's runs folder. Creating new run in the config-based runs folder."
                )
                checkpoints_save_folder = os.path.join(
                    config.save_folder, agent_controller_config.run_name + run_suffix
                )
        else:
            checkpoints_save_folder = os.path.join(
                config.save_folder, agent_controller_config.run_name + run_suffix
            )
        self.checkpoints_save_folder = checkpoints_save_folder
        print(
            f"{config.agent_controller_name}: Saving checkpoints to {self.checkpoints_save_folder}"
        )

        self.learner.load(
            DerivedPPOLearnerConfig(
                learner_config=learner_config,
                agent_controller_name=config.agent_controller_name,
                obs_space=self.obs_space,
                action_space=self.action_space,
                checkpoint_load_folder=learner_checkpoint_load_folder,
            )
        )
        self.experience_buffer.load(
            DerivedExperienceBufferConfig(
                experience_buffer_config=experience_buffer_config,
                agent_controller_name=config.agent_controller_name,
                seed=config.base_config.random_seed,
                dtype=agent_controller_config.learner_config.dtype,
                learner_device=agent_controller_config.learner_config.device,
                checkpoint_load_folder=experience_buffer_checkpoint_load_folder,
            )
        )
        if self.metrics_logger is not None:
            metrics_logger_config = self.metrics_logger.validate_config(
                self.config.agent_controller_config.metrics_logger_config
            )
            if isinstance(self.metrics_logger, WandbMetricsLogger):
                additional_derived_config = WandbAdditionalDerivedConfig(
                    derived_wandb_run_config={
                        **self.config.agent_controller_config.learner_config.model_dump(),
                        "exp_buffer_size": self.config.agent_controller_config.experience_buffer_config.max_size,
                        "timesteps_per_iteration": self.config.agent_controller_config.timesteps_per_iteration,
                        "n_proc": self.config.process_config.n_proc,
                        "min_process_steps_per_inference": self.config.process_config.min_process_steps_per_inference,
                        "timestep_limit": self.config.base_config.timestep_limit,
                        **self.config.agent_controller_config.experience_buffer_config.trajectory_processor_config,
                    },
                    timestamp_suffix=run_suffix,
                )
            else:
                additional_derived_config = None
            self.metrics_logger.load(
                DerivedMetricsLoggerConfig(
                    metrics_logger_config=metrics_logger_config,
                    checkpoint_load_folder=metrics_logger_checkpoint_load_folder,
                    agent_controller_name=config.agent_controller_name,
                    additional_derived_config=additional_derived_config,
                )
            )

        if agent_controller_config.checkpoint_load_folder is not None:
            self._load_from_checkpoint()

        torch.manual_seed(self.config.base_config.random_seed)
        np.random.seed(self.config.base_config.random_seed)
        random.seed(self.config.base_config.random_seed)

    def _load_from_checkpoint(self):
        try:
            with open(
                os.path.join(
                    self.config.agent_controller_config.checkpoint_load_folder,
                    ITERATION_TRAJECTORIES_FILE,
                ),
                "rb",
            ) as f:
                iteration_trajectories: List[
                    Trajectory[AgentID, ObsType, ActionType, RewardType]
                ] = pickle.load(f)
        except FileNotFoundError:
            print(
                f"{self.config.agent_controller_name}: Tried to load current trajectories from checkpoint using the file at location {str(os.path.join(self.config.agent_controller_config.checkpoint_load_folder, ITERATION_TRAJECTORIES_FILE))}, but there is no such file! Current trajectories will be initialized as an empty list instead."
            )
            iteration_trajectories = []
        try:
            with open(
                os.path.join(
                    self.config.agent_controller_config.checkpoint_load_folder,
                    ITERATION_SHARED_INFOS_FILE,
                ),
                "rb",
            ) as f:
                iteration_shared_infos: List[Dict[str, Any]] = pickle.load(f)
        except FileNotFoundError:
            print(
                f"{self.config.agent_controller_name}: Tried to load iteration shared info data from checkpoint using the file at location {str(os.path.join(self.config.agent_controller_config.checkpoint_load_folder, ITERATION_SHARED_INFOS_FILE))}, but there is no such file! Iteration shared info data will be initialized as an empty list instead."
            )
            iteration_shared_infos = []
        try:
            with open(
                os.path.join(
                    self.config.agent_controller_config.checkpoint_load_folder,
                    PPO_AGENT_FILE,
                ),
                "rt",
            ) as f:
                state = json.load(f)
        except FileNotFoundError:
            print(
                f"{self.config.agent_controller_name}: Tried to load PPO agent miscellaneous state data from checkpoint using the file at location {str(os.path.join(self.config.agent_controller_config.checkpoint_load_folder, PPO_AGENT_FILE))}, but there is no such file! This state data will be initialized as if this were a new run instead."
            )
            state = {
                "cur_iteration": 0,
                "iteration_timesteps": 0,
                "cumulative_timesteps": 0,
                "iteration_start_time": time.perf_counter(),
                "timestep_collection_start_time": time.perf_counter(),
            }

        self.iteration_trajectories = iteration_trajectories
        self.iteration_shared_infos = iteration_shared_infos
        self.cur_iteration = state["cur_iteration"]
        self.iteration_timesteps = state["iteration_timesteps"]
        self.cumulative_timesteps = state["cumulative_timesteps"]
        # I'm aware that loading these start times will cause some funny numbers for the first iteration
        self.iteration_start_time = state["iteration_start_time"]
        self.timestep_collection_start_time = state["timestep_collection_start_time"]

    def save_checkpoint(self):
        print(f"Saving checkpoint {self.cumulative_timesteps}...")

        checkpoint_save_folder = os.path.join(
            self.checkpoints_save_folder, str(time.time_ns())
        )
        os.makedirs(checkpoint_save_folder, exist_ok=True)
        self.learner.save_checkpoint(
            os.path.join(checkpoint_save_folder, PPO_LEARNER_FOLDER)
        )
        self.experience_buffer.save_checkpoint(
            os.path.join(checkpoint_save_folder, EXPERIENCE_BUFFER_FOLDER)
        )
        if self.metrics_logger is not None:
            self.metrics_logger.save_checkpoint(
                os.path.join(checkpoint_save_folder, METRICS_LOGGER_FOLDER)
            )

        if self.config.agent_controller_config.save_mid_iteration_data_in_checkpoint:
            with open(
                os.path.join(checkpoint_save_folder, ITERATION_TRAJECTORIES_FILE),
                "wb",
            ) as f:
                pickle.dump(self.iteration_trajectories, f)
            with open(
                os.path.join(checkpoint_save_folder, ITERATION_SHARED_INFOS_FILE),
                "wb",
            ) as f:
                pickle.dump(self.iteration_shared_infos, f)
        with open(os.path.join(checkpoint_save_folder, PPO_AGENT_FILE), "wt") as f:
            state = {
                "cur_iteration": self.cur_iteration,
                "iteration_timesteps": self.iteration_timesteps,
                "cumulative_timesteps": self.cumulative_timesteps,
                "iteration_start_time": self.iteration_start_time,
                "timestep_collection_start_time": self.timestep_collection_start_time,
            }
            json.dump(state, f, indent=4)

        # TODO: does this actually work? I'm not sure the file structure I'm using actually works with this assumption
        # Prune old checkpoints
        existing_checkpoints = [
            int(arg) for arg in os.listdir(self.checkpoints_save_folder)
        ]
        if (
            len(existing_checkpoints)
            > self.config.agent_controller_config.n_checkpoints_to_keep
        ):
            existing_checkpoints.sort()
            for checkpoint_name in existing_checkpoints[
                : -self.config.agent_controller_config.n_checkpoints_to_keep
            ]:
                shutil.rmtree(
                    os.path.join(self.checkpoints_save_folder, str(checkpoint_name))
                )

    def choose_agents(self, agent_id_list):
        return self.agent_choice_fn(agent_id_list)

    @torch.no_grad
    def get_actions(self, agent_id_list, obs_list):
        action_list, log_probs = self.learner.actor.get_action(agent_id_list, obs_list)
        if log_probs.dim() == 0:
            # This can happen if the input is a single element
            log_probs = log_probs.unsqueeze(0)
        return (action_list, log_probs)

    def standardize_timestep_observations(
        self,
        timesteps: List[Timestep[AgentID, ObsType, ActionType, RewardType]],
    ):
        agent_id_list = [None] * (2 * len(timesteps))
        obs_list = [None] * len(agent_id_list)
        for timestep_idx, timestep in enumerate(timesteps):
            agent_id_list[2 * timestep_idx] = timestep.agent_id
            agent_id_list[2 * timestep_idx + 1] = timestep.agent_id
            obs_list[2 * timestep_idx] = timestep.obs
            obs_list[2 * timestep_idx + 1] = timestep.next_obs
        standardized_obs = self.obs_standardizer.standardize(agent_id_list, obs_list)
        for obs_idx, obs in enumerate(standardized_obs):
            if obs_idx % 2 == 0:
                timesteps[obs_idx // 2].obs = obs
            else:
                timesteps[obs_idx // 2].next_obs = obs

    def process_timestep_data(self, timestep_data):
        timesteps_added = 0
        shared_infos: List[Dict[str, Any]] = []
        for env_id, (
            env_timesteps,
            env_log_probs,
            env_shared_info,
            _,
        ) in timestep_data.items():
            if self.obs_standardizer is not None:
                self.standardize_timestep_observations(env_timesteps)
            if env_timesteps:
                if env_id not in self.current_env_trajectories:
                    self.current_env_trajectories[env_id] = EnvTrajectories(
                        [timestep.agent_id for timestep in env_timesteps],
                        self.agent_choice_fn,
                    )
                timesteps_added += self.current_env_trajectories[env_id].add_steps(
                    env_timesteps, env_log_probs
                )
            shared_infos.append(env_shared_info)
        self.iteration_timesteps += timesteps_added
        self.cumulative_timesteps += timesteps_added
        self.iteration_shared_infos += shared_infos
        if (
            self.iteration_timesteps
            >= self.config.agent_controller_config.timesteps_per_iteration
        ):
            self.timestep_collection_end_time = time.perf_counter()
            self._learn()
            self.cur_iteration += 1
        if self.ts_since_last_save >= self.config.agent_controller_config.save_every_ts:
            self.save_checkpoint()
            self.ts_since_last_save = 0

    def choose_env_actions(self, state_info):
        env_action_responses = {}
        for env_id in state_info:
            if env_id not in self.current_env_trajectories:
                # This must be the first env action after a reset, so we step
                env_action_responses[env_id] = EnvActionResponse.STEP()
                continue
            done = all(self.current_env_trajectories[env_id].dones.values())
            if done:
                env_action_responses[env_id] = EnvActionResponse.RESET()
            else:
                env_action_responses[env_id] = EnvActionResponse.STEP()
        return env_action_responses

    def process_env_actions(self, env_actions):
        for env_id, env_action in env_actions.items():
            # this is a getter so we only want to do it once
            enum_type = env_action.enum_type
            if enum_type == EnvActionResponseType.STEP:
                pass
            elif enum_type == EnvActionResponseType.RESET:
                env_trajectories = self.current_env_trajectories.pop(env_id)
                env_trajectories.finalize()
                self.iteration_trajectories += env_trajectories.get_trajectories()
            elif enum_type == EnvActionResponseType.SET_STATE:
                # Can get the desired_state using env_action.desired_state and the prev_timestep_id_dict using env_action.prev_timestep_id_dict, but I'll leave that to you
                raise NotImplementedError
            else:
                raise ValueError

    def _learn(self):
        env_trajectories_list = list(self.current_env_trajectories.values())
        for env_trajectories in env_trajectories_list:
            env_trajectories.finalize()
            self.iteration_trajectories += env_trajectories.get_trajectories()
        self._update_value_predictions()
        trajectory_processor_data = self.experience_buffer.submit_experience(
            self.iteration_trajectories
        )
        ppo_data = self.learner.learn(self.experience_buffer)

        cur_time = time.perf_counter()
        if self.metrics_logger is not None:
            self.metrics_logger.collect_agent_metrics(
                PPOAgentControllerData(
                    ppo_data,
                    trajectory_processor_data,
                    self.cumulative_timesteps,
                    cur_time - self.iteration_start_time,
                    self.iteration_timesteps,
                    self.timestep_collection_end_time
                    - self.timestep_collection_start_time,
                )
            )
            self.metrics_logger.collect_env_metrics(self.iteration_shared_infos)
            self.metrics_logger.report_metrics()

        self.iteration_trajectories.clear()
        self.iteration_shared_infos.clear()
        self.current_env_trajectories.clear()
        self.ts_since_last_save += self.iteration_timesteps
        self.iteration_timesteps = 0
        self.iteration_start_time = cur_time
        self.timestep_collection_start_time = time.perf_counter()

    @torch.no_grad()
    def _update_value_predictions(self):
        """
        Function to update the value predictions inside the Trajectory instances of self.iteration_trajectories
        """
        traj_timestep_idx_ranges: List[Tuple[int, int]] = []
        start = 0
        stop = 0
        critic_agent_id_input: List[AgentID] = []
        critic_obs_input: List[ObsType] = []
        for trajectory in self.iteration_trajectories:
            obs_list = trajectory.obs_list + [trajectory.final_obs]
            traj_len = len(obs_list)
            agent_id_list = [trajectory.agent_id] * traj_len
            stop = start + traj_len
            critic_agent_id_input += agent_id_list
            critic_obs_input += obs_list
            traj_timestep_idx_ranges.append((start, stop))
            start = stop

        val_preds: torch.Tensor = (
            self.learner.critic(critic_agent_id_input, critic_obs_input)
            .flatten()
            .to(device="cpu", non_blocking=True)
        )
        torch.cuda.empty_cache()
        for idx, (start, stop) in enumerate(traj_timestep_idx_ranges):
            self.iteration_trajectories[idx].val_preds = val_preds[start : stop - 1]
            self.iteration_trajectories[idx].final_val_pred = val_preds[stop - 1]
        if self.config.agent_controller_config.learner_config.device.type != "cpu":
            torch.cuda.current_stream().synchronize()



"""
File: torch_functions.py
Author: Matthew Allen

Description:
    A helper file for misc. PyTorch functions.

"""

import torch
import torch.nn as nn


def get_device(device: str):
    if device in ["auto", "gpu"] and torch.cuda.is_available():
        return "cuda:0"
    elif device == "auto" and not torch.cuda.is_available():
        return "cpu"
    else:
        return device


class MapContinuousToAction(nn.Module):
    """
    A class for policies using the continuous action space. Continuous policies output N*2 values for N actions where
    each value is in the range [-1, 1]. Half of these values will be used as the mean of a multi-variate normal distribution
    and the other half will be used as the diagonal of the covariance matrix for that distribution. Since variance must
    be positive, this class will map the range [-1, 1] for those values to the desired range (defaults to [0.1, 1]) using
    a simple linear transform.
    """

    def __init__(self, range_min=0.1, range_max=1):
        super().__init__()

        tanh_range = [-1, 1]
        self.m = (range_max - range_min) / (tanh_range[1] - tanh_range[0])
        self.b = range_min - tanh_range[0] * self.m

    def forward(self, x):
        n = x.shape[-1] // 2
        # map the right half of x from [-1, 1] to [range_min, range_max].
        return x[..., :n], x[..., n:] * self.m + self.b


class MultiDiscreteRolv(nn.Module):
    """
    A class to handle the multi-discrete action space in Rocket League. There are 8 potential actions, 5 of which can be
    any of {-1, 0, 1} and 3 of which can be either of {0, 1}. This class takes 21 logits, appends -inf to the final 3
    such that each of the 8 actions has 3 options (to avoid a ragged list), then builds a categorical distribution over
    each class for each action. Credit to Rolv Arild for coming up with this method.
    """

    def __init__(self, bins):
        super().__init__()
        self.distribution = None
        self.bins = bins

    def make_distribution(self, logits):
        """
        Function to make the multi-discrete categorical distribution for a group of logits.
        :param logits: Logits which parameterize the distribution.
        :return: None.
        """

        # Split the 21 logits into the expected bins.
        logits = torch.split(logits, self.bins, dim=-1)

        # Separate triplets from the split logits.
        triplets = torch.stack(logits[:5], dim=-1)

        # Separate duets and pad the final dimension with -inf to create triplets.
        duets = torch.nn.functional.pad(
            torch.stack(logits[5:], dim=-1), pad=(0, 0, 0, 1), value=float("-inf")
        )

        # Un-split the logits now that the duets have been converted into triplets and reshape them into the correct shape.
        logits = torch.cat((triplets, duets), dim=-1).swapdims(-1, -2)

        # Construct a distribution with our fixed logits.
        self.distribution = torch.distributions.Categorical(logits=logits)

    def log_prob(self, action) -> torch.Tensor:
        return self.distribution.log_prob(action).sum(dim=-1)

    def sample(self):
        return self.distribution.sample()

    def entropy(self):
        return self.distribution.entropy().sum(
            dim=-1
        )  # Unsure about this sum operation.



from .batch_reward_type_numpy_converter import (
    BatchRewardTypeNumpyConverter,
    BatchRewardTypeSimpleNumpyConverter,
)
from .numpy_obs_standardizer import NumpyObsStandardizer
from .obs_standardizer import ObsStandardizer



import json
import os
import time
from typing import Optional


def convert_rlgym_ppo_checkpoint(
    rlgym_ppo_checkpoint_folder: str, out_folder: Optional[str]
):

    if out_folder is None:
        out_folder = f"rlgym_ppo_converted_checkpoint_{time.time_ns()}"
    print(f"Saving converted checkpoint to folder {out_folder}")

    os.makedirs(out_folder, exist_ok=True)

    PPO_FILES = [
        ("PPO_POLICY_OPTIMIZER.pt", "actor_optimizer.pt"),
        ("PPO_POLICY.pt", "actor.pt"),
        ("PPO_VALUE_NET_OPTIMIZER.pt", "critic_optimizer.pt"),
        ("PPO_VALUE_NET.pt", "critic.pt"),
    ]
    os.makedirs(f"{out_folder}/ppo_learner", exist_ok=True)
    for file in PPO_FILES:
        with open(f"{rlgym_ppo_checkpoint_folder}/{file[0]}", "rb") as fin:
            with open(f"{out_folder}/ppo_learner/{file[1]}", "wb") as fout:
                fout.write(fin.read())



import os
import pickle
from dataclasses import dataclass
from typing import Any, Dict, Generic, Iterable, List, Optional, Tuple

import numpy as np
import torch
from pydantic import BaseModel, Field, model_validator
from rlgym.api import ActionType, AgentID, ObsType, RewardType

from rlgym_learn_algos.util.torch_functions import get_device
from rlgym_learn_algos.util.torch_pydantic import PydanticTorchDevice

from .trajectory import Trajectory
from .trajectory_processor import (
    DerivedTrajectoryProcessorConfig,
    TrajectoryProcessor,
    TrajectoryProcessorConfig,
    TrajectoryProcessorData,
)

EXPERIENCE_BUFFER_FILE = "experience_buffer.pkl"


class ExperienceBufferConfigModel(BaseModel, extra="forbid"):
    max_size: int = 100000
    device: PydanticTorchDevice = "auto"
    save_experience_buffer_in_checkpoint: bool = True
    trajectory_processor_config: Dict[str, Any] = Field(default_factory=dict)

    @model_validator(mode="before")
    @classmethod
    def set_trajectory_processor_config(cls, data):
        if isinstance(data, ExperienceBufferConfigModel):
            if isinstance(data.trajectory_processor_config, BaseModel):
                data.trajectory_processor_config = (
                    data.trajectory_processor_config.model_dump()
                )
        elif isinstance(data, dict):
            if "trajectory_processor_config" in data:
                if isinstance(data["trajectory_processor_config"], BaseModel):
                    data["trajectory_processor_config"] = data[
                        "trajectory_processor_config"
                    ].model_dump()
            if "device" not in data:
                data["device"] = "auto"
            data["device"] = get_device(data["device"])
        return data


@dataclass
class DerivedExperienceBufferConfig:
    experience_buffer_config: ExperienceBufferConfigModel
    agent_controller_name: str
    seed: int
    dtype: torch.dtype
    learner_device: torch.device
    checkpoint_load_folder: Optional[str] = None


class ExperienceBuffer(
    Generic[
        TrajectoryProcessorConfig,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        TrajectoryProcessorData,
    ]
):
    @staticmethod
    def _cat(t1, t2, size):
        t2_len = len(t2)
        if t2_len > size:
            # t2 alone is larger than we want; copy the end
            # This clone is needed to avoid nesting views
            t = t2[-size:].clone()

        elif t2_len == size:
            # t2 is a perfect match; just use it directly
            t = t2

        elif len(t1) + t2_len > size:
            # t1+t2 is larger than we want; use t2 wholly with the end of t1 before it
            t = torch.cat((t1[t2_len - size :], t2), 0)

        else:
            # t1+t2 does not exceed what we want; concatenate directly
            t = torch.cat((t1, t2), 0)

        del t1
        del t2
        return t

    @staticmethod
    def _cat_list(cur, new, size):
        new_len = len(new)
        if new_len > size:
            t = new[-size:]
        elif new_len == size:
            t = new
        elif len(cur) + new_len > size:
            t = cur[new_len - size :] + new
        else:
            t = cur + new
        return t

    def __init__(
        self,
        trajectory_processor: TrajectoryProcessor[
            TrajectoryProcessorConfig,
            AgentID,
            ObsType,
            ActionType,
            RewardType,
            TrajectoryProcessorData,
        ],
    ):
        self.trajectory_processor = trajectory_processor
        self.agent_ids: List[AgentID] = []
        self.observations: List[ObsType] = []
        self.actions: List[ActionType] = []

    def load(self, config: DerivedExperienceBufferConfig):
        self.config = config
        self.rng = np.random.RandomState(config.seed)
        trajectory_processor_config = self.trajectory_processor.validate_config(
            config.experience_buffer_config.trajectory_processor_config
        )
        self.trajectory_processor.load(
            DerivedTrajectoryProcessorConfig(
                trajectory_processor_config=trajectory_processor_config,
                agent_controller_name=config.agent_controller_name,
                dtype=config.dtype,
                device=config.learner_device,
            )
        )
        self.log_probs = torch.tensor([], dtype=config.dtype)
        self.values = torch.tensor([], dtype=config.dtype)
        self.advantages = torch.tensor([], dtype=config.dtype)
        if self.config.checkpoint_load_folder is not None:
            self._load_from_checkpoint()
        self.log_probs = self.log_probs.to(config.learner_device)
        self.values = self.values.to(config.learner_device)
        self.advantages = self.advantages.to(config.learner_device)

    def _load_from_checkpoint(self):
        # lazy way
        # TODO: don't use pickle for torch things, use torch.load because of map_location. Or maybe define a custom unpickler for this? Or maybe one already exists?
        try:
            with open(
                os.path.join(
                    self.config.checkpoint_load_folder, EXPERIENCE_BUFFER_FILE
                ),
                "rb",
            ) as f:
                state_dict = pickle.load(f)
            self.agent_ids = state_dict["agent_ids"]
            self.observations = state_dict["observations"]
            self.actions = state_dict["actions"]
            self.log_probs = state_dict["log_probs"]
            self.values = state_dict["values"]
            self.advantages = state_dict["advantages"]
        except FileNotFoundError:
            print(
                f"{self.config.agent_controller_name}: Tried to load experience buffer from checkpoint using the file at location {str(os.path.join(self.config.checkpoint_load_folder, EXPERIENCE_BUFFER_FILE))}, but there is no such file! A blank experience buffer will be used instead."
            )

    def save_checkpoint(self, folder_path):
        os.makedirs(folder_path, exist_ok=True)
        if self.config.experience_buffer_config.save_experience_buffer_in_checkpoint:
            with open(
                os.path.join(folder_path, EXPERIENCE_BUFFER_FILE),
                "wb",
            ) as f:
                pickle.dump(
                    {
                        "agent_ids": self.agent_ids,
                        "observations": self.observations,
                        "actions": self.actions,
                        "log_probs": self.log_probs,
                        "values": self.values,
                        "advantages": self.advantages,
                    },
                    f,
                )
        self.trajectory_processor.save_checkpoint(folder_path)

    # TODO: update docs
    def submit_experience(
        self, trajectories: List[Trajectory[AgentID, ActionType, ObsType, RewardType]]
    ) -> TrajectoryProcessorData:
        """
        Function to add experience to the buffer.

        :param observations: An ordered sequence of observations from the environment.
        :param actions: The corresponding actions that were taken at each state in the `states` sequence.
        :param log_probs: The log probability for each action in `actions`
        :param rewards: A list of rewards such that rewards[i] is the reward for taking action actions[i] from observation observations[i]
        :param terminateds: An ordered sequence of the terminated flags from the environment.
        :param truncateds: An ordered sequence of the truncated flag from the environment.
        :param values: The output of the value function estimator evaluated on the observations.
        :param advantages: The advantage of each action at each state in `states` and `actions`

        :return: TrajectoryProcessorData
        """

        _cat = ExperienceBuffer._cat
        _cat_list = ExperienceBuffer._cat_list
        exp_buffer_data, trajectory_processor_data = (
            self.trajectory_processor.process_trajectories(trajectories)
        )
        (agent_ids, observations, actions, log_probs, values, advantages) = (
            exp_buffer_data
        )

        self.agent_ids = _cat_list(
            self.agent_ids, agent_ids, self.config.experience_buffer_config.max_size
        )
        self.observations = _cat_list(
            self.observations,
            observations,
            self.config.experience_buffer_config.max_size,
        )
        self.actions = _cat_list(
            self.actions, actions, self.config.experience_buffer_config.max_size
        )
        self.log_probs = _cat(
            self.log_probs,
            log_probs,
            self.config.experience_buffer_config.max_size,
        )
        self.values = _cat(
            self.values,
            values,
            self.config.experience_buffer_config.max_size,
        )
        self.advantages = _cat(
            self.advantages,
            advantages,
            self.config.experience_buffer_config.max_size,
        )

        return trajectory_processor_data

    # TODO: tensordict?
    def _get_samples(self, indices) -> Tuple[
        Iterable[AgentID],
        Iterable[ObsType],
        Iterable[ActionType],
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
    ]:
        return (
            [self.agent_ids[index] for index in indices],
            [self.observations[index] for index in indices],
            [self.actions[index] for index in indices],
            self.log_probs[indices],
            self.values[indices],
            self.advantages[indices],
        )

    def get_all_batches_shuffled(self, batch_size):
        """
        Function to return the experience buffer in shuffled batches. Code taken from the stable-baeselines3 buffer:
        https://github.com/DLR-RM/stable-baselines3/blob/2ddf015cd9840a2a1675f5208be6eb2e86e4d045/stable_baselines3/common/buffers.py#L482
        :param batch_size: size of each batch yielded by the generator.
        :return:
        """
        if self.config.learner_device.type != "cpu":
            torch.cuda.current_stream().synchronize()
        total_samples = self.values.shape[0]
        indices = self.rng.permutation(total_samples)
        start_idx = 0
        while start_idx + batch_size <= total_samples:
            yield self._get_samples(indices[start_idx : start_idx + batch_size])
            start_idx += batch_size

    def clear(self):
        """
        Function to clear the experience buffer.
        :return: None.
        """
        del self.observations
        del self.actions
        del self.log_probs
        del self.values
        del self.advantages
        self.__init__(self.trajectory_processor)



"""
File: continuous_policy.py
Author: Matthew Allen

Description:
    An implementation of a policy for the continuous action space. For N actions the policy will output N*2 values in
    the range [-1, 1]. A linear transform is applied to the latter half of the policy's output to map it onto the range
    [var_min, var_max]. A mutli-variate Gaussian distribution with a diagonal covariance matrix is then constructed
    using the former half of the policy's output as the mean and the transformed latter half as the diagonal of the
    covariance matrix.

"""

import functools
from typing import Iterable, List, Tuple

import numpy as np
import torch
import torch.nn as nn
from rlgym.api import AgentID
from torch.distributions import Normal

from rlgym_learn_algos.util import torch_functions

from .actor import Actor


class ContinuousActor(Actor[AgentID, np.ndarray, np.ndarray]):
    def __init__(
        self, input_shape, output_shape, layer_sizes, device, var_min=0.1, var_max=1.0
    ):
        super().__init__()
        self.device = device
        self.affine_map = torch_functions.MapContinuousToAction(
            range_min=var_min, range_max=var_max
        )

        # Build the neural network.
        assert (
            len(layer_sizes) != 0
        ), "AT LEAST ONE LAYER MUST BE SPECIFIED TO BUILD THE NEURAL NETWORK!"
        layers = [nn.Linear(input_shape, layer_sizes[0]), nn.ReLU()]

        prev_size = layer_sizes[0]
        for size in layer_sizes[1:]:
            layers.append(nn.Linear(prev_size, size))
            layers.append(nn.ReLU())
            prev_size = size

        layers.append(nn.Linear(layer_sizes[-1], output_shape))
        layers.append(nn.Tanh())
        self.model = nn.Sequential(*layers).to(self.device)

    @functools.lru_cache()
    def logpdf(self, x, mean, std):
        """
        Function to compute the log of the pdf of our distribution parameterized by (mean, std) evaluated at x. PyTorch
        can do this natively but I don't trust their method.
        :param x: value to compute the logpdf for.
        :param mean: Mean of the distribution to evaluate.
        :param std: Diagonal of the covariance matrix of the distribution to evaluate.
        :return: ln(pdf(x)).
        """

        msq = mean * mean
        ssq = std * std
        xsq = x * x

        term1 = -torch.divide(msq, (2 * ssq))
        term2 = torch.divide(mean * x, ssq)
        term3 = -torch.divide(xsq, (2 * ssq))
        term4 = torch.log(1 / torch.sqrt(2 * np.pi * ssq))

        return term1 + term2 + term3 + term4

    def get_output(
        self, obs_list: List[np.ndarray]
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        obs = torch.as_tensor(
            np.array(obs_list), dtype=torch.float32, device=self.device
        )

        policy_output = self.model(obs)
        return self.affine_map(policy_output)

    def get_action(
        self, agent_id_list, obs_list, **kwargs
    ) -> Tuple[Iterable[np.ndarray], torch.Tensor]:
        mean, std = self.get_output(obs_list)
        if "deterministic" in kwargs and kwargs["deterministic"]:
            # The probability of a deterministic action occurring is 1 -> log(1) = 0.
            return mean.cpu().numpy(), torch.zeros(mean.shape)

        distribution = Normal(loc=mean, scale=std)
        action = distribution.sample().clamp(min=-1, max=1)
        log_prob = self.logpdf(action, mean, std)

        shape = log_prob.shape
        if "summed_probs" in kwargs and kwargs["summed_probs"]:
            if len(shape) > 1:
                log_prob = log_prob.sum(dim=-1)
            else:
                log_prob = log_prob.sum()

        return action.cpu().numpy(), log_prob.cpu().squeeze()

    def get_backprop_data(self, agent_id_list, obs_list, acts, **kwargs):
        mean, std = self.get_output(obs_list)
        distribution = Normal(loc=mean, scale=std)

        acts_tensor = torch.as_tensor(np.array(acts)).to(self.device)
        prob = self.logpdf(acts_tensor, mean, std)
        if kwargs["summed_probs"]:
            log_probs = prob.sum(dim=1).to(self.device)
        else:
            log_probs = prob.to(self.device)

        entropy = distribution.entropy()
        entropy = entropy.mean().to(self.device)

        return log_probs, entropy



from dataclasses import dataclass
from typing import Generic, List

from rlgym.api import ActionType, AgentID, ObsType, RewardType
from torch import Tensor


@dataclass
class Trajectory(Generic[AgentID, ObsType, ActionType, RewardType]):
    __slots__ = (
        "agent_id",
        "obs_list",
        "action_list",
        "log_probs",
        "reward_list",
        "val_preds",
        "final_obs",
        "final_val_pred",
        "truncated",
    )
    agent_id: AgentID
    obs_list: List[ObsType]
    action_list: List[ActionType]
    log_probs: Tensor
    reward_list: List[RewardType]
    val_preds: Tensor
    final_obs: ObsType
    final_val_pred: Tensor
    truncated: bool



import json
import os
from typing import List

import numpy as np
import torch
from rlgym.api import ActionType, AgentID, ObsType, RewardType

from rlgym_learn_algos.stateful_functions import (
    BatchRewardTypeNumpyConverter,
    BatchRewardTypeSimpleNumpyConverter,
)
from rlgym_learn_algos.util.running_stats import WelfordRunningStat

from .gae_trajectory_processor import (
    GAETrajectoryProcessorConfigModel,
    GAETrajectoryProcessorData,
)
from .trajectory_processor import TRAJECTORY_PROCESSOR_FILE, TrajectoryProcessor


class GAETrajectoryProcessorPurePython(
    TrajectoryProcessor[
        GAETrajectoryProcessorConfigModel,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        GAETrajectoryProcessorData,
    ]
):
    def __init__(
        self,
        batch_reward_type_numpy_converter: BatchRewardTypeNumpyConverter[
            RewardType
        ] = BatchRewardTypeSimpleNumpyConverter(),
    ):
        """
        :param gamma: Gamma hyper-parameter.
        :param lmbda: Lambda hyper-parameter.
        :param return_std: Standard deviation of the returns (used for reward normalization).
        """
        self.return_stats = WelfordRunningStat(1)
        self.batch_reward_type_numpy_converter = batch_reward_type_numpy_converter

    def process_trajectories(self, trajectories):
        return_std = (
            self.return_stats.std.squeeze() if self.standardize_returns else None
        )
        gamma = np.array(self.gamma, dtype=self.dtype)
        lmbda = np.array(self.lmbda, dtype=self.dtype)
        exp_len = 0
        agent_ids: List[AgentID] = []
        observations: List[ObsType] = []
        actions: List[ActionType] = []
        # For some reason, appending to lists is faster than preallocating the tensor and then indexing into it to assign
        log_probs_list: List[torch.Tensor] = []
        values_list: List[torch.Tensor] = []
        advantages_list: List[torch.Tensor] = []
        returns_list: List[torch.Tensor] = []
        reward_sum = np.array(0, dtype=self.dtype)
        for trajectory in trajectories:
            cur_return = np.array(0, dtype=self.dtype)
            next_val_pred = (
                trajectory.final_val_pred.squeeze().cpu().numpy()
                if trajectory.truncated
                else np.array(0, dtype=self.dtype)
            )

            cur_advantages = np.array(0, dtype=self.dtype)
            reward_array = self.batch_reward_type_numpy_converter.as_numpy(
                trajectory.reward_list
            )
            value_preds = trajectory.val_preds.unbind(0)
            for obs, action, log_prob, reward, value_pred in reversed(
                list(
                    zip(
                        trajectory.obs_list,
                        trajectory.action_list,
                        trajectory.log_probs,
                        np.nditer(reward_array),
                        value_preds,
                    )
                )
            ):
                val_pred = value_pred.cpu().numpy()
                reward_sum += reward
                if return_std is not None:
                    norm_reward = np.clip(
                        reward / return_std,
                        a_min=self.norm_reward_min,
                        a_max=self.norm_reward_max,
                    )
                else:
                    norm_reward = reward
                delta = norm_reward + gamma * next_val_pred - val_pred
                next_val_pred = val_pred
                cur_advantages = delta + gamma * lmbda * cur_advantages
                cur_return = reward + gamma * cur_return
                returns_list.append(cur_return)
                agent_ids.append(trajectory.agent_id)
                observations.append(obs)
                actions.append(action)
                log_probs_list.append(log_prob)
                values_list.append(value_pred)
                advantages_list.append(cur_advantages)
                exp_len += 1

        if self.standardize_returns:
            # Update the running statistics about the returns.
            n_to_increment = min(self.max_returns_per_stats_increment, exp_len)

            for sample in returns_list[:n_to_increment]:
                self.return_stats.update(sample)
            avg_return = self.return_stats.mean
            return_std = self.return_stats.std
        else:
            avg_return = np.nan
            return_std = np.nan
        avg_reward = reward_sum.item() / exp_len
        average_episode_return = reward_sum.item() / len(trajectories)
        trajectory_processor_data = GAETrajectoryProcessorData(
            average_undiscounted_episodic_return=average_episode_return,
            average_return=avg_return,
            return_standard_deviation=return_std,
            average_reward=avg_reward
        )
        return (
            (
                agent_ids,
                observations,
                actions,
                torch.stack(log_probs_list).to(device=self.device),
                torch.stack(values_list).to(device=self.device),
                torch.from_numpy(np.array(advantages_list)).to(device=self.device),
            ),
            trajectory_processor_data,
        )

    def validate_config(self, config_obj):
        return GAETrajectoryProcessorConfigModel.model_validate(config_obj)

    def load(self, config):
        self.gamma = config.trajectory_processor_config.gamma
        self.lmbda = config.trajectory_processor_config.lmbda
        self.standardize_returns = (
            config.trajectory_processor_config.standardize_returns
        )
        self.max_returns_per_stats_increment = (
            config.trajectory_processor_config.max_returns_per_stats_increment
        )
        self.dtype = np.dtype(str(config.dtype).replace("torch.", ""))
        self.device = config.device
        self.checkpoint_load_folder = config.checkpoint_load_folder
        if self.checkpoint_load_folder is not None:
            self._load_from_checkpoint()
        self.norm_reward_min = np.array(-10, dtype=self.dtype)
        self.norm_reward_max = np.array(10, dtype=self.dtype)
        self.batch_reward_type_numpy_converter.set_dtype(self.dtype)

    def _load_from_checkpoint(self):
        with open(
            os.path.join(self.checkpoint_load_folder, TRAJECTORY_PROCESSOR_FILE),
            "rt",
        ) as f:
            state = json.load(f)
        self.return_stats.load_state_dict(state["return_running_stats"])

    def save_checkpoint(self, folder_path):
        state = {
            "return_running_stats": self.return_stats.state_dict(),
        }
        with open(
            os.path.join(folder_path, TRAJECTORY_PROCESSOR_FILE),
            "wt",
        ) as f:
            json.dump(state, f, indent=4)



from .rlgym_learn_algos import (
    DerivedGAETrajectoryProcessorConfig as RustDerivedGAETrajectoryProcessorConfig,
)
from .rlgym_learn_algos import GAETrajectoryProcessor as RustGAETrajectoryProcessor



from abc import abstractmethod
from typing import Any, Dict, Generic, List

from rlgym_learn.api import AgentControllerData

from .metrics_logger import (
    DerivedMetricsLoggerConfig,
    MetricsLoggerAdditionalDerivedConfig,
    MetricsLoggerConfig,
)


def print_dict(d: dict, indent=""):
    deferred_list = []
    for k, v in d.items():
        if isinstance(k, str):
            k_str = k
        else:
            k_str = repr(k)
        if isinstance(v, dict):
            deferred_list.append((k_str, v))
            continue
        if isinstance(v, str):
            v_str = v
        else:
            v_str = repr(v)
        print(f"{indent}{k_str}: {v_str}")
    for k_str, v in deferred_list:
        print(f"{indent}-- {k_str} --")
        print_dict(v, indent + "  ")


class DictMetricsLogger(
    Generic[
        MetricsLoggerConfig,
        MetricsLoggerAdditionalDerivedConfig,
        AgentControllerData,
    ]
):
    """
    This is a specification of the MetricsLogger which provides an additional method get_metrics to retrieve the metrics as a dictionary.
    """

    def collect_env_metrics(self, data: List[Dict[str, Any]]):
        pass

    def collect_agent_metrics(self, data: AgentControllerData):
        pass

    @abstractmethod
    def get_metrics(self) -> Dict[str, Any]:
        """
        :return: metrics data for consumption and side effects by the caller, in the form of a dictionary
        """
        raise NotImplementedError

    def report_metrics(self):
        print_dict(self.get_metrics())

    @abstractmethod
    def validate_config(self, config_obj: Dict[str, Any]) -> MetricsLoggerConfig:
        raise NotImplementedError

    def load(
        self,
        config: DerivedMetricsLoggerConfig[
            MetricsLoggerConfig, MetricsLoggerAdditionalDerivedConfig
        ],
    ):
        pass

    def save_checkpoint(self, folder_path):
        pass



"""
File: multi_discrete_policy.py
Author: Matthew Allen

Description:
    An implementation of a feed-forward neural network which parametrizes 8 discrete distributions over the actions
    available in Rocket League.
"""

from typing import Iterable, List, Tuple

import numpy as np
import torch
import torch.nn as nn
from rlgym.api import AgentID

from rlgym_learn_algos.util import torch_functions

from .actor import Actor


class MultiDiscreteFF(Actor[AgentID, np.ndarray, np.ndarray]):
    def __init__(self, input_shape, layer_sizes, device):
        super().__init__()
        self.device = device
        bins = [3, 3, 3, 3, 3, 2, 2, 2]
        n_output_nodes = sum(bins)
        assert (
            len(layer_sizes) != 0
        ), "AT LEAST ONE LAYER MUST BE SPECIFIED TO BUILD THE NEURAL NETWORK!"
        layers = [nn.Linear(input_shape, layer_sizes[0]), nn.ReLU()]

        prev_size = layer_sizes[0]
        for size in layer_sizes[1:]:
            layers.append(nn.Linear(prev_size, size))
            layers.append(nn.ReLU())
            prev_size = size

        layers.append(nn.Linear(layer_sizes[-1], n_output_nodes))
        self.model = nn.Sequential(*layers).to(self.device)
        self.splits = bins
        self.multi_discrete = torch_functions.MultiDiscreteRolv(bins)

    def get_output(self, obs_list: List[np.ndarray]):
        obs = torch.as_tensor(
            np.array(obs_list), dtype=torch.float32, device=self.device
        )
        policy_output = self.model(obs)
        return policy_output

    def get_action(
        self, agent_id_list, obs_list, **kwargs
    ) -> Tuple[Iterable[np.ndarray], torch.Tensor]:
        logits = self.get_output(obs_list)

        # TODO not sure how to do this better - very slow atm
        if "deterministic" in kwargs and kwargs["deterministic"]:
            start = 0
            action = []
            for split in self.splits:
                action.append(logits[..., start : start + split].argmax(dim=-1))
                start += split
            action = torch.stack(action).cpu().numpy()
            return action, 0

        distribution = self.multi_discrete
        distribution.make_distribution(logits)

        action = distribution.sample()
        log_prob = distribution.log_prob(action)

        # returned action shape: (L, 8) where L is the batching dimension (parallel with obs_list)
        # Returned log prob shape: (L) where L is the batching dimension (parallel with obs_list)
        return action.cpu().numpy(), log_prob.cpu()

    def get_backprop_data(
        self, agent_id_list, obs_list, acts, **kwargs
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        logits = self.get_output(obs_list)
        acts_tensor = torch.as_tensor(np.array(acts)).to(self.device)

        distribution = self.multi_discrete
        distribution.make_distribution(logits)

        entropy = distribution.entropy().to(self.device)
        log_probs = distribution.log_prob(acts_tensor).to(self.device)

        return log_probs, entropy.mean()



from __future__ import annotations

import json
import os
from dataclasses import dataclass

import numpy as np
import torch
from numpy import ndarray
from pydantic import BaseModel
from rlgym.api import ActionType, AgentID, ObsType, RewardType

from rlgym_learn_algos.stateful_functions import (
    BatchRewardTypeNumpyConverter,
    BatchRewardTypeSimpleNumpyConverter,
)
from rlgym_learn_algos.util.running_stats import WelfordRunningStat

from ..ppo import RustDerivedGAETrajectoryProcessorConfig, RustGAETrajectoryProcessor
from .trajectory_processor import TRAJECTORY_PROCESSOR_FILE, TrajectoryProcessor


class GAETrajectoryProcessorConfigModel(BaseModel, extra="forbid"):
    gamma: float = 0.99
    lmbda: float = 0.95
    standardize_returns: bool = True
    max_returns_per_stats_increment: int = 150


@dataclass
class GAETrajectoryProcessorData:
    average_reward: float
    average_undiscounted_episodic_return: float
    average_return: float
    return_standard_deviation: float


class GAETrajectoryProcessor(
    TrajectoryProcessor[
        GAETrajectoryProcessorConfigModel,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        GAETrajectoryProcessorData,
    ]
):
    def __init__(
        self,
        batch_reward_type_numpy_converter: BatchRewardTypeNumpyConverter[
            RewardType
        ] = BatchRewardTypeSimpleNumpyConverter(),
    ):
        """
        :param batch_reward_type_numpy_converter: BatchRewardTypeNumpyConverter instance
        """
        self.return_stats = WelfordRunningStat(1)
        self.rust_gae_trajectory_processor: RustGAETrajectoryProcessor[
            AgentID, ObsType, ActionType, RewardType
        ] = RustGAETrajectoryProcessor(
            batch_reward_type_numpy_converter,
        )

    def process_trajectories(self, trajectories):
        return_std = self.return_stats.std[0] if self.standardize_returns else 1
        (
            agent_id_list,
            observation_list,
            action_list,
            log_probs,
            value_preds,
            advantage_array,
            return_array,
            avg_reward,
            avg_undiscounted_return,
        ) = self.rust_gae_trajectory_processor.process_trajectories(
            trajectories, return_std
        )

        if self.standardize_returns:
            # Update the running statistics about the returns.
            n_to_increment = min(
                self.max_returns_per_stats_increment, len(return_array)
            )

            for sample in return_array[:n_to_increment]:
                self.return_stats.update(sample)
        trajectory_processor_data = GAETrajectoryProcessorData(
            average_reward=avg_reward,
            average_undiscounted_episodic_return=avg_undiscounted_return,
            average_return=return_array.mean(),
            return_standard_deviation=return_array.std(),
        )
        return (
            (
                agent_id_list,
                observation_list,
                action_list,
                log_probs.to(device=self.device),
                value_preds.to(device=self.device),
                torch.from_numpy(advantage_array).to(device=self.device),
            ),
            trajectory_processor_data,
        )

    def validate_config(self, config_obj):
        return GAETrajectoryProcessorConfigModel.model_validate(config_obj)

    def load(self, config):
        self.gamma = config.trajectory_processor_config.gamma
        self.lmbda = config.trajectory_processor_config.lmbda
        self.standardize_returns = (
            config.trajectory_processor_config.standardize_returns
        )
        self.max_returns_per_stats_increment = (
            config.trajectory_processor_config.max_returns_per_stats_increment
        )
        self.agent_controller_name = config.agent_controller_name
        self.dtype = config.dtype
        self.device = config.device
        self.checkpoint_load_folder = config.checkpoint_load_folder
        if self.checkpoint_load_folder is not None:
            self._load_from_checkpoint()
        self.rust_gae_trajectory_processor.load(
            RustDerivedGAETrajectoryProcessorConfig(
                self.gamma, self.lmbda, np.dtype(str(self.dtype)[6:])
            )
        )

    def _load_from_checkpoint(self):
        try:
            with open(
                os.path.join(self.checkpoint_load_folder, TRAJECTORY_PROCESSOR_FILE),
                "rt",
            ) as f:
                state = json.load(f)
            self.return_stats.load_state_dict(state["return_running_stats"])
        except FileNotFoundError:
            print(
                f"{self.agent_controller_name}: Tried to load trajectory processor from checkpoint using the trajectory processor file at location {str(os.path.join(self.checkpoint_load_folder, TRAJECTORY_PROCESSOR_FILE))}, but there is no such file! Running stats will be initialized as if this were a new run instead."
            )

    def save_checkpoint(self, folder_path):
        state = {
            "return_running_stats": self.return_stats.state_dict(),
        }
        with open(
            os.path.join(folder_path, TRAJECTORY_PROCESSOR_FILE),
            "wt",
        ) as f:
            json.dump(state, f, indent=4)



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, Generic, List, Optional, Tuple, TypeVar

from rlgym.api import ActionType, AgentID, ObsType, RewardType
from torch import Tensor, device, dtype

from .trajectory import Trajectory

TrajectoryProcessorConfig = TypeVar("TrajectoryProcessorConfig")
TrajectoryProcessorData = TypeVar("TrajectoryProcessorData")

TRAJECTORY_PROCESSOR_FILE = "trajectory_processor.json"


@dataclass
class DerivedTrajectoryProcessorConfig(Generic[TrajectoryProcessorConfig]):
    trajectory_processor_config: TrajectoryProcessorConfig
    agent_controller_name: str
    dtype: dtype
    device: device
    checkpoint_load_folder: Optional[str] = None


class TrajectoryProcessor(
    Generic[
        TrajectoryProcessorConfig,
        AgentID,
        ObsType,
        ActionType,
        RewardType,
        TrajectoryProcessorData,
    ]
):
    @abstractmethod
    def process_trajectories(
        self,
        trajectories: List[Trajectory[AgentID, ActionType, ObsType, RewardType]],
    ) -> Tuple[
        Tuple[List[AgentID], List[ObsType], List[ActionType], Tensor, Tensor, Tensor],
        TrajectoryProcessorData,
    ]:
        """
        :param trajectories: List of Trajectory instances from which to generate experience.
        :return: Tuple of (Tuple of parallel lists (considering tensors as a list in their first dimension)
            with (AgentID, ObsType), ActionType, log prob, value, and advantage respectively) and
            TrajectoryProcessorData (for use in the MetricsLogger).
            log prob, value, and advantage tensors should be with dtype=dtype and device=device.
        """
        raise NotImplementedError

    @abstractmethod
    def validate_config(self, config_obj: Dict[str, Any]) -> TrajectoryProcessorConfig:
        raise NotImplementedError

    def load(self, config: DerivedTrajectoryProcessorConfig[TrajectoryProcessorConfig]):
        pass

    def save_checkpoint(self, folder_path):
        pass



from typing import Callable, Dict, Generic, List, Optional

import torch
from rlgym.api import ActionType, AgentID, ObsType, RewardType
from rlgym_learn import Timestep
from torch import Tensor

from .trajectory import Trajectory


class EnvTrajectories(Generic[AgentID, ObsType, ActionType, RewardType]):
    def __init__(
        self,
        agent_ids: List[AgentID],
        agent_choice_fn: Callable[
            [List[AgentID]], List[int]
        ] = lambda agent_id_list: list(range(len(agent_id_list))),
    ) -> None:
        self.used_agent_id_idx_map = {
            agent_ids[idx]: idx for idx in agent_choice_fn(agent_ids)
        }
        self.obs_lists: Dict[AgentID, List[ObsType]] = {}
        self.action_lists: Dict[AgentID, List[ActionType]] = {}
        self.reward_lists: Dict[AgentID, List[RewardType]] = {}
        self.final_obs: Dict[AgentID, Optional[ObsType]] = {}
        self.dones: Dict[AgentID, bool] = {}
        self.truncateds: Dict[AgentID, bool] = {}
        for agent_id in self.used_agent_id_idx_map:
            self.obs_lists[agent_id] = []
            self.action_lists[agent_id] = []
            self.reward_lists[agent_id] = []
            self.final_obs[agent_id] = None
            self.dones[agent_id] = False
            self.truncateds[agent_id] = False
        self.log_probs_list = []

    def add_steps(self, timesteps: List[Timestep], log_probs: Tensor):
        steps_added = 0
        for timestep in timesteps:
            agent_id = timestep.agent_id
            # We only want to process the timesteps of agent ids we included from this env when creating the EnvTrajectories instance
            if agent_id not in self.used_agent_id_idx_map:
                continue
            if not self.dones[agent_id]:
                steps_added += 1
                self.obs_lists[agent_id].append(timestep.obs)
                self.action_lists[agent_id].append(timestep.action)
                self.reward_lists[agent_id].append(timestep.reward)
                self.final_obs[agent_id] = timestep.next_obs
                now_done = timestep.terminated or timestep.truncated
                if now_done:
                    self.dones[agent_id] = True
                    self.truncateds[agent_id] = timestep.truncated
        # We append all the log probs but we will deal with this later when getting trajectories
        self.log_probs_list.append(log_probs)
        return steps_added

    def finalize(self):
        """
        Truncates any unfinished trajectories, marks all trajectories as done.
        """
        for agent_id in self.used_agent_id_idx_map:
            self.truncateds[agent_id] = (
                self.truncateds[agent_id] or not self.dones[agent_id]
            )
            self.dones[agent_id] = True

    def get_trajectories(
        self,
    ) -> List[Trajectory[AgentID, ObsType, ActionType, RewardType]]:
        """
        :return: List of trajectories relevant to this env
        """
        log_probs = torch.stack(self.log_probs_list)
        trajectories = []
        for agent_id, idx in self.used_agent_id_idx_map.items():
            obs_list = self.obs_lists[agent_id]
            trajectories.append(
                Trajectory(
                    agent_id,
                    obs_list,
                    self.action_lists[agent_id],
                    log_probs[: len(obs_list), idx],
                    self.reward_lists[agent_id],
                    None,
                    self.final_obs[agent_id],
                    torch.tensor(0, dtype=torch.float32),
                    self.truncateds[agent_id],
                )
            )
        return trajectories



from rlgym_learn_algos import (
    RustDerivedGAETrajectoryProcessorConfig,
    RustGAETrajectoryProcessor,
)

from .actor import Actor
from .basic_critic import BasicCritic
from .continuous_actor import ContinuousActor
from .critic import Critic
from .discrete_actor import DiscreteFF
from .experience_buffer import (
    DerivedExperienceBufferConfig,
    ExperienceBuffer,
    ExperienceBufferConfigModel,
)
from .experience_buffer_numpy import NumpyExperienceBuffer
from .gae_trajectory_processor import (
    GAETrajectoryProcessor,
    GAETrajectoryProcessorConfigModel,
    GAETrajectoryProcessorData,
)
from .gae_trajectory_processor_pure_python import GAETrajectoryProcessorPurePython
from .multi_discrete_actor import MultiDiscreteFF
from .ppo_agent_controller import (
    PPOAgentController,
    PPOAgentControllerConfigModel,
    PPOAgentControllerData,
)
from .ppo_learner import (
    DerivedPPOLearnerConfig,
    PPOData,
    PPOLearner,
    PPOLearnerConfigModel,
)
from .ppo_metrics_logger import PPOMetricsLogger
from .trajectory import Trajectory
from .trajectory_processor import TrajectoryProcessor, TrajectoryProcessorData



from typing import List, Optional

import numpy as np
import torch
from rlgym.api import AgentID, RewardType

from .experience_buffer import ExperienceBuffer
from .trajectory_processor import TrajectoryProcessorConfig, TrajectoryProcessorData


class NumpyExperienceBuffer(
    ExperienceBuffer[
        TrajectoryProcessorConfig,
        AgentID,
        np.ndarray,
        np.ndarray,
        RewardType,
        TrajectoryProcessorData,
    ],
):

    @staticmethod
    def _cat_numpy(t1, t2, size):
        t2 = np.array(t2)
        t2_len = len(t2)
        if t1 is None:
            if t2_len > size:
                t = t2[-size:].copy()
            else:
                t = t2
            del t1
            del t2
            return t

        if t2_len > size:
            # t2 alone is larger than we want; copy the end
            # This clone is needed to avoid nesting views
            t = t2[-size:].copy()

        elif t2_len == size:
            # t2 is a perfect match; just use it directly
            t = t2

        elif len(t1) + t2_len > size:
            # t1+t2 is larger than we want; use t2 wholly with the end of t1 before it
            t = np.concatenate((t1[t2_len - size :], t2), 0)

        else:
            # t1+t2 does not exceed what we want; concatenate directly
            t = np.concatenate((t1, t2), 0)

        del t1
        del t2
        return t

    def __init__(
        self,
        trajectory_processor,
    ):
        self.trajectory_processor = trajectory_processor
        self.agent_ids: List[AgentID] = []
        self.observations: Optional[np.ndarray] = None
        self.actions: Optional[np.ndarray] = None
        self.log_probs = torch.FloatTensor()
        self.values = torch.FloatTensor()
        self.advantages = torch.FloatTensor()

    def submit_experience(self, trajectories):
        _cat = ExperienceBuffer._cat
        _cat_list = ExperienceBuffer._cat_list
        _cat_numpy = NumpyExperienceBuffer._cat_numpy
        exp_buffer_data, trajectory_processor_data = (
            self.trajectory_processor.process_trajectories(trajectories)
        )
        (agent_ids, observations, actions, log_probs, values, advantages) = (
            exp_buffer_data
        )

        self.agent_ids = _cat_list(
            self.agent_ids, agent_ids, self.config.experience_buffer_config.max_size
        )
        self.observations = _cat_numpy(
            self.observations,
            observations,
            self.config.experience_buffer_config.max_size,
        )
        self.actions = _cat_numpy(
            self.actions, actions, self.config.experience_buffer_config.max_size
        )
        self.log_probs = _cat(
            self.log_probs,
            log_probs,
            self.config.experience_buffer_config.max_size,
        )
        self.values = _cat(
            self.values,
            values,
            self.config.experience_buffer_config.max_size,
        )
        self.advantages = _cat(
            self.advantages,
            advantages,
            self.config.experience_buffer_config.max_size,
        )

        return trajectory_processor_data

    def _get_samples(self, indices):
        return (
            [self.agent_ids[index] for index in indices],
            self.observations[indices],
            self.actions[indices],
            self.log_probs[indices],
            self.values[indices],
            self.advantages[indices],
        )

    def get_all_batches_shuffled(self, batch_size):
        """
        Function to return the experience buffer in shuffled batches. Code taken from the stable-baeselines3 buffer:
        https://github.com/DLR-RM/stable-baselines3/blob/2ddf015cd9840a2a1675f5208be6eb2e86e4d045/stable_baselines3/common/buffers.py#L482
        :param batch_size: size of each batch yielded by the generator.
        :return:
        """
        if self.config.experience_buffer_config.device.type != "cpu":
            torch.cuda.current_stream().synchronize()
        total_samples = self.values.shape[0]
        indices = self.rng.permutation(total_samples)
        start_idx = 0
        while start_idx + batch_size <= total_samples:
            yield self._get_samples(indices[start_idx : start_idx + batch_size])
            start_idx += batch_size

    def clear(self):
        """
        Function to clear the experience buffer.
        :return: None.
        """
        del self.observations
        del self.actions
        del self.log_probs
        del self.values
        del self.advantages
        self.__init__(self.max_size, self.seed, self.device)



from typing import Annotated, Any

import torch
from pydantic import (
    BaseModel,
    GetCoreSchemaHandler,
    GetJsonSchemaHandler,
    ValidationError,
)
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

dtype_str_regex = "|".join(
    set(
        f"({str(v)[6:]})" for v in torch.__dict__.values() if isinstance(v, torch.dtype)
    )
)
device_str_regex = (
    "("
    + "|".join(
        f"({v})"
        for v in [
            "cpu",
            "cuda",
            "ipu",
            "xpu",
            "mkldnn",
            "opengl",
            "opencl",
            "ideep",
            "hip",
            "ve",
            "fpga",
            "maia",
            "xla",
            "lazy",
            "vulkan",
            "mps",
            "meta",
            "hpu",
            "mtia",
            "privateuseone",
        ]
    )
    + ")(:\d+)?"
)


# Created using the example here: https://docs.pydantic.dev/latest/concepts/types/#handling-third-party-types
class _TorchDtypePydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(pattern=dtype_str_regex),
                core_schema.no_info_plain_validator_function(
                    lambda v: getattr(torch, v)
                ),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(torch.dtype),
                    from_str_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: str(v)[6:]
            ),
        )


class _TorchDevicePydanticAnnotation:
    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source_type: Any,
        _handler: GetCoreSchemaHandler,
    ) -> core_schema.CoreSchema:
        from_str_schema = core_schema.chain_schema(
            [
                core_schema.str_schema(pattern=device_str_regex),
                core_schema.no_info_plain_validator_function(lambda v: torch.device(v)),
            ]
        )
        from_int_schema = core_schema.chain_schema(
            [
                core_schema.int_schema(ge=0),
                core_schema.no_info_plain_validator_function(lambda v: torch.device(v)),
            ]
        )

        return core_schema.json_or_python_schema(
            json_schema=from_str_schema,
            python_schema=core_schema.union_schema(
                [
                    # check if it's an instance first before doing any further work
                    core_schema.is_instance_schema(torch.dtype),
                    from_str_schema,
                    from_int_schema,
                ]
            ),
            serialization=core_schema.plain_serializer_function_ser_schema(
                lambda v: str(v)
            ),
        )


PydanticTorchDtype = Annotated[torch.dtype, _TorchDtypePydanticAnnotation]
PydanticTorchDevice = Annotated[torch.device, _TorchDevicePydanticAnnotation]



from __future__ import annotations

from typing import TYPE_CHECKING, Generic, List, Tuple

from numpy import dtype, ndarray
from rlgym.api import ActionType, AgentID, ObsType, RewardType

from rlgym_learn_algos.stateful_functions import BatchRewardTypeNumpyConverter

if TYPE_CHECKING:
    from torch import Tensor

    from rlgym_learn_algos.ppo import Trajectory

class DerivedGAETrajectoryProcessorConfig:
    def __new__(
        cls, gamma: float, lmbda: float, dtype: dtype
    ) -> DerivedGAETrajectoryProcessorConfig: ...

class GAETrajectoryProcessor(Generic[AgentID, ObsType, ActionType, RewardType]):
    def __new__(
        cls, batch_reward_type_numpy_converter: BatchRewardTypeNumpyConverter
    ) -> GAETrajectoryProcessor: ...
    def load(self, config: DerivedGAETrajectoryProcessorConfig): ...
    def process_trajectories(
        self,
        trajectories: List[Trajectory[AgentID, ObsType, ActionType, RewardType]],
        return_std: ndarray,
    ) -> Tuple[
        List[AgentID],
        List[ObsType],
        List[ActionType],
        Tensor,
        Tensor,
        ndarray,
        ndarray,
        float,
        float,
    ]: ...
